"""
This module contains Okta integration steps.
"""

from __future__ import annotations

from typing import Dict, Optional, Union
from logging import Filter

from requests import HTTPError

from koheesio.logger import LoggingFactory, MaskedString
from koheesio.models import Field, SecretStr, model_validator
from koheesio.steps.http import HttpPostStep


class Okta(HttpPostStep):
    """
    Base Okta class
    """

    client_id: str = Field(default=..., alias="okta_id", description="Okta account ID")
    client_secret: SecretStr = Field(default=..., alias="okta_secret", description="Okta account secret", repr=False)
    data: Optional[Union[Dict[str, str], str]] = Field(
        default={"grant_type": "client_credentials"}, description="Data to be sent along with the token request"
    )

    @model_validator(mode="before")
    def _set_auth_param(cls, v):
        """
        Assign auth parameter with Okta client and secret to the params dictionary.
        If auth parameter already exists, it will be overwritten.
        """
        auth = (v["client_id"], MaskedString(v["client_secret"].get_secret_value()))
        v["params"] = {"auth": auth} if not v.get("params") else {**v["params"], "auth": auth}
        return v


class LoggerOktaTokenFilter(Filter):
    """Filter which hides token value from log."""

    def __init__(self, okta_object: OktaAccessToken, name: str = "OktaToken"):
        self.__okta_object = okta_object
        super().__init__(name=name)

    def filter(self, record):
        # noinspection PyUnresolvedReferences
        if token := self.__okta_object.output.token:
            token_value = token.get_secret_value()
            record.msg = record.msg.replace(token_value, "<SECRET_TOKEN>")

        return True


class OktaAccessToken(Okta):
    """
    Get Okta authorization token

    Example:
    ```python
    token = (
        OktaAccessToken(
            url="https://org.okta.com",
            client_id="client",
            client_secret=SecretStr("secret"),
            params={
                "p1": "foo",
                "p2": "bar",
            },
        )
        .execute()
        .token
    )
    ```
    """

    class Output(Okta.Output):
        """Output class for OktaAccessToken."""

        token: Optional[SecretStr] = Field(default=None, description="Okta authentication token")

    def __init__(self, **kwargs):
        _logger = LoggingFactory.get_logger(name=self.__class__.__name__, inherit_from_koheesio=True)
        logger_filter = LoggerOktaTokenFilter(okta_object=self)
        _logger.addFilter(logger_filter)
        super().__init__(**kwargs)

    def execute(self):
        """
        Execute an HTTP Post call to Okta service and retrieve the access token.
        """
        HttpPostStep.execute(self)

        # noinspection PyUnresolvedReferences
        status_code = self.output.status_code
        # noinspection PyUnresolvedReferences
        raw_payload = self.output.raw_payload

        if status_code != 200:
            raise HTTPError(f"Request failed with '{status_code}' code. Payload: {raw_payload}")

        # noinspection PyUnresolvedReferences
        json_payload = self.output.json_payload

        if token := json_payload.get("access_token"):
            self.output.token = SecretStr(token)
        else:
            raise ValueError(f"No 'access_token' found in the Okta response: {json_payload}")



"""Base class for a Pandas step

Extends the Step class with Pandas DataFrame support. The following:
- Pandas steps are expected to return a Pandas DataFrame as output.
"""

from typing import Optional
from abc import ABC

from koheesio import Step, StepOutput
from koheesio.models import Field
from koheesio.spark.utils import import_pandas_based_on_pyspark_version

pandas = import_pandas_based_on_pyspark_version()


class PandasStep(Step, ABC):
    """Base class for a Pandas step

    Extends the Step class with Pandas DataFrame support. The following:
    - Pandas steps are expected to return a Pandas DataFrame as output.
    """

    class Output(StepOutput):
        """Output class for PandasStep"""

        df: Optional[pandas.DataFrame] = Field(default=None, description="The Pandas DataFrame")



import pytest
from chispa import assert_df_equality

from pyspark.sql.types import StructType

from koheesio.spark.readers.memory import DataFormat, InMemoryDataReader

pytestmark = pytest.mark.spark


class TestInMemoryDataReader:
    # fmt: off
    @pytest.mark.parametrize(
        "data,format,params,expect_filter",
        [
            pytest.param(
                "id,string\n1,hello,\n2,world", DataFormat.CSV, {"header": True}, "id < 3"
            ),
            pytest.param(
                b"id,string\n1,hello,\n2,world", DataFormat.CSV, {"header": True}, "id < 3"
            ),
            pytest.param(
                '{"id": 1, "string": "hello"}', DataFormat.JSON, {}, "id < 2"
            ),
            pytest.param(
                {"id": 1, "string": "hello"}, DataFormat.JSON, {}, "id < 2"
            ),
            pytest.param(
                ['{"id": 1, "string": "hello"}', '{"id": 2, "string": "world"}'], DataFormat.JSON, {}, "id < 3"
            )
        ],
    )
    # fmt: on
    def test_execute(self, spark, sample_df_with_strings, data, format, params, expect_filter):
        df = (
            InMemoryDataReader(
                data=data,
                format=format,
                schema_=StructType.fromJson(
                    {
                        "fields": [
                            {"metadata": {}, "name": "id", "nullable": True, "type": "long"},
                            {"metadata": {}, "name": "string", "nullable": True, "type": "string"},
                        ],
                        "type": "struct",
                    }
                ),
                params=params,
            )
            .execute()
            .df
        )
        assert_df_equality(df1=df, df2=sample_df_with_strings.where(expect_filter))



import pytest

from pyspark.sql import SparkSession

from koheesio.spark.transformations.lookup import (
    DataframeLookup,
    JoinHint,
    JoinMapping,
    JoinType,
    TargetColumn,
)

pytestmark = pytest.mark.spark


def test_join_mapping_column_test() -> None:
    assert str(JoinMapping(source_column="source", other_column="joined").column) == "Column<'joined AS source'>"


def test_target_column_column_test() -> None:
    assert str(TargetColumn(target_column="target", target_column_alias="alias").column) == "Column<'target AS alias'>"


def test_join_type_values() -> None:
    assert JoinType.INNER == "inner"
    assert JoinType.FULL == "full"
    assert JoinType.LEFT == "left"
    assert JoinType.RIGHT == "right"
    assert JoinType.CROSS == "cross"
    assert JoinType.SEMI == "semi"
    assert JoinType.ANTI == "anti"


def test_join_hint_values() -> None:
    assert JoinHint.BROADCAST == "broadcast"
    assert JoinHint.MERGE == "merge"


@pytest.mark.parametrize("join_hint", [None, JoinHint.BROADCAST])
def test_dataframe_lookup(spark: SparkSession, join_hint: JoinHint) -> None:
    df = spark.createDataFrame(
        [("1", "a", "a"), ("2", "b", "b")],
        schema="key string, second_key string, field string",
    )
    right_df = spark.createDataFrame(
        [("1", "a", "aa", "aaa", ""), ("3", "c", "cc", "ccc", "")],
        schema="other_key string, other_second_key string, bad_field string, worse_field string, useless_field string",
    )
    assert (
        DataframeLookup(
            df=df,
            other=right_df,
            on=[
                JoinMapping(source_column="key", other_column="other_key"),
                JoinMapping(source_column="second_key", other_column="other_second_key"),
            ],
            targets=[
                TargetColumn(target_column="bad_field", target_column_alias="good_field"),
                TargetColumn(target_column="worse_field", target_column_alias="better_field"),
            ],
            how=JoinType.INNER,
            hint=join_hint,
        )
        .execute()
        .df.head()
        .asDict()
    ) == {
        "key": "1",
        "second_key": "a",
        "field": "a",
        "good_field": "aa",
        "better_field": "aaa",
    }



from collections import namedtuple
from io import BytesIO
from pathlib import PurePath

import pytest

from pydantic import ValidationError

from koheesio.integrations.box import (
    Box,
    BoxCsvFileReader,
    BoxCsvPathReader,
    BoxFileBase,
    BoxFileWriter,
    BoxFolderBase,
    BoxFolderCreate,
    BoxFolderDelete,
    BoxFolderGet,
    BoxFolderNotFoundError,
    BoxPathIsEmptyError,
    File,
    Folder,
    JWTAuth,
    SecretStr,
    StructType,
)

pytestmark = pytest.mark.spark

COMMON_PARAMS = {
    "client_id": SecretStr("client_id"),
    "client_secret": SecretStr("client_secret"),
    "enterprise_id": SecretStr("enterprise_id"),
    "jwt_key_id": SecretStr("jwt_key_id"),
    "rsa_private_key_data": SecretStr("rsa_private_key_data"),
    "rsa_private_key_passphrase": SecretStr("rsa_private_key_passphrase"),
}
CUSTOM_SCHEMA = StructType.fromJson(
    {
        "fields": [
            {"metadata": {}, "name": "foo", "nullable": True, "type": "string"},
            {"metadata": {}, "name": "bar", "nullable": True, "type": "integer"},
        ],
        "type": "struct",
    }
)


@pytest.fixture(autouse=True)
def dummy_box(mocker):
    class DummyBox:
        def __init__(self, *args, **kwargs):
            self.args = args

        def _mock_folder(self, name: str, object_id: str, get_items: list):
            mf = mocker.Mock(spec=Folder)
            mf.object_id = object_id
            mf.name = name
            mf.type = "folder"
            mf.get_items.return_value = get_items
            mf.create_subfolder.return_value = mocker.Mock(type="folder", object_id="999", spec=Folder)
            mf.upload_stream.side_effect = self._mock_file
            return mf

        @staticmethod
        def _mock_file(name: str = None, object_id: str = "0", content: bytes = None, **kwargs):
            mf = mocker.Mock(spec=File)
            mf.object_id = object_id
            mf.name = name if name else kwargs["file_name"]
            mf.type = "file"
            mf.content.return_value = content
            Properties = namedtuple("Properties", "name")
            mf.get.return_value = Properties(name=mf.name)
            mf.get_shared_link.return_value = "https://my-ent.box.com/dummy"
            return mf

        def folder(self, *args, **kwargs):
            """
            Mock the following folder and file structure:
                root/
                ├─ bar/
                ├─ foo/
                │  ├─ f1.csv
                │  ├─ f2.csv
                │  ├─ f3.xlsx
            """
            mock_foo = self._mock_folder(
                name="foo",
                object_id="1",
                get_items=[
                    self._mock_file(name="f1.csv", object_id="10"),
                    self._mock_file(name="f2.csv", object_id="20"),
                    self._mock_file(name="f3.xlsx", object_id="30"),
                ],
            )
            mock_bar = self._mock_folder(
                name="bar",
                object_id="2",
                get_items=[],
            )

            mock_root = self._mock_folder(name="root", object_id="0", get_items=[mock_foo, mock_bar])

            if kwargs["folder_id"] == "0":
                return mock_root

            if kwargs["folder_id"] == "1":
                return mock_foo

        def file(self, *args, **kwargs):
            if kwargs["file_id"] == "10":
                return self._mock_file(name="f1.csv", object_id="10", content=b"foo,bar\nA,1")
            if kwargs["file_id"] == "20":
                return self._mock_file(name="f2.csv", object_id="20", content=b"foo,bar\nA,2")
            if kwargs["file_id"] == "30":
                return self._mock_file(name="upload.csv", object_id="30", content=b"foo,bar\nA,2")

    mocker.patch("koheesio.integrations.box.JWTAuth", spec=JWTAuth)
    mocker.patch("koheesio.integrations.box.Client", new=DummyBox)


class TestBox:
    def test__init__(self, dummy_box):
        b = Box(**COMMON_PARAMS)
        assert "JWTAuth" in str(b.client.args[0])


class TestBoxFolderBase:
    def test_validate_folder_or_path_w_both(self):
        with pytest.raises(ValidationError):
            BoxFolderBase(folder="0", path="foo")

    def test_validate_folder_or_path_wo_both(self):
        with pytest.raises(ValidationError):
            BoxFolderBase()

    def test_action(self):
        with pytest.raises(NotImplementedError):
            BoxFolderBase(**COMMON_PARAMS, folder="0").action()


class TestBoxFolderGet:
    def test_get_w_default_root(self, dummy_box):
        folder = (
            BoxFolderGet(
                **COMMON_PARAMS,
                path="/foo",
                root="0",
            )
            .execute()
            .folder
        )
        assert folder.name == "foo"
        assert len(folder.get_items()) == 3

    def test_get_w_folder_and_path(self, dummy_box):
        with pytest.raises(AttributeError):
            BoxFolderGet(
                **COMMON_PARAMS,
                folder="1",
                path="/foo",
            ).execute().folder

    def test_get_w_custom_root(self, dummy_box):
        folder = (
            BoxFolderGet(
                **COMMON_PARAMS,
                path="/",
                root="1",
            )
            .execute()
            .folder
        )
        assert folder.name == "foo"
        assert len(folder.get_items()) == 3

    def test_get_w_invalid_path(self, dummy_box):
        with pytest.raises(BoxFolderNotFoundError):
            (
                BoxFolderGet(
                    **COMMON_PARAMS,
                    path="/buz",
                    root="0",
                )
                .execute()
                .folder
            )


class TestBoxFolderCreate:
    def test_create_w_path(
        self,
        dummy_box,
    ):
        folder = (
            BoxFolderCreate(
                **COMMON_PARAMS,
                path="/foo/buz",
            )
            .execute()
            .folder
        )
        assert folder.object_id == "999"

    def test_create_w_folder(self, dummy_box):
        with pytest.raises(AttributeError):
            BoxFolderCreate(**COMMON_PARAMS, folder="0")


class TestBoxFolderDelete:
    def test_execute(self, dummy_box):
        BoxFolderDelete(**COMMON_PARAMS, path="/foo/buz")


class TestBoxFileBase:
    def test_execute(self, mocker):
        action = mocker.patch.object(BoxFileBase, "action")
        BoxFileBase(**COMMON_PARAMS, files=["10", "20"], path="/foo").execute()
        assert action.call_count == 2
        assert action.call_args_list[0].kwargs["file"].name == "f1.csv"
        assert action.call_args_list[0].kwargs["folder"].name == "foo"


class TestBoxCsvReader:
    def test_execute_w_schema(self, spark, dummy_box):
        bcr = BoxCsvFileReader(
            **COMMON_PARAMS,
            file=["10"],
            schema=CUSTOM_SCHEMA,
        ).execute()
        assert bcr.df.count() == 1
        assert bcr.df.dtypes == [
            ("foo", "string"),
            ("bar", "int"),
            ("meta_file_id", "string"),
            ("meta_file_name", "string"),
            ("meta_load_timestamp", "timestamp"),
        ]

    def test_execute_wo_schema(self, spark, dummy_box):
        bcr = BoxCsvFileReader(
            **COMMON_PARAMS,
            file=["10"],
        ).execute()
        assert bcr.df.count() == 1
        assert bcr.df.dtypes == [
            ("foo", "string"),
            ("bar", "string"),
            ("meta_file_id", "string"),
            ("meta_file_name", "string"),
            ("meta_load_timestamp", "timestamp"),
        ]


class TestBoxCsvPathReader:
    def test_execute_w_files(self, spark, dummy_box):
        bcr = BoxCsvPathReader(
            **COMMON_PARAMS,
            path="/foo",
            schema=CUSTOM_SCHEMA,
        ).execute()
        assert bcr.df.count() == 2
        assert bcr.df.dtypes == [
            ("foo", "string"),
            ("bar", "int"),
            ("meta_file_id", "string"),
            ("meta_file_name", "string"),
            ("meta_load_timestamp", "timestamp"),
        ]

    def test_execute_wo_files(self, spark, dummy_box):
        with pytest.raises(BoxPathIsEmptyError):
            BoxCsvPathReader(
                **COMMON_PARAMS,
                path="/bar",
            ).execute()


class TestBoxFileWriter:
    def test_execute_bytesio_wo_file_name(self, dummy_box):
        with pytest.raises(AttributeError):
            BoxFileWriter(**COMMON_PARAMS, path="/foo", file=BytesIO(b"my-file")).execute()

    def test_execute_bytesio(self, dummy_box):
        f = BoxFileWriter(
            **COMMON_PARAMS,
            path="/foo",
            file=BytesIO(b"my-data"),
            file_name="upload.txt",
        ).execute()
        assert f.shared_link
        assert isinstance(f.file, File)
        assert f.file.get().name == "upload.txt"

    def test_execute_file_wo_file_name(self, dummy_box):
        # file_name not set, will default to original file name
        f = BoxFileWriter(**COMMON_PARAMS, path="/foo", file=__file__).execute()
        assert f.file.get().name == PurePath(__file__).name

    def test_execute_file_w_file_name(self, dummy_box):
        # file_name set, will override original file name
        f = BoxFileWriter(
            **COMMON_PARAMS,
            path="/foo",
            file=__file__,
            file_name="upload.txt",
        ).execute()
        assert f.file.get().name == "upload.txt"



import json
from typing import Optional
from textwrap import dedent

import pytest
import yaml

from koheesio.context import Context
from koheesio.models import BaseModel, ExtraParamsMixin


class TestBaseModel:
    class SimpleModel(BaseModel):
        a: int
        b: str = "default"

    class FooModel(BaseModel):
        foo: Optional[str] = None
        baz: Optional[int] = None

    def test_a_simple_model(self):
        """Test a simple model."""
        foo = self.SimpleModel(a=1)
        assert foo.model_dump() == {"a": 1, "b": "default", "description": "SimpleModel", "name": "SimpleModel"}

    def test_context_management_no_exception(self):
        with self.SimpleModel.lazy() as m:
            m.a = 1
            m.b = "test"
        assert m.a == 1
        assert m.b == "test"

    def test_context_management_with_exception(self):
        """The context manager should raise the original exception after exiting the context."""
        with pytest.raises(ValueError):
            with self.SimpleModel.lazy() as m:
                m.a = 1
                m.b = "test"
                raise ValueError("Test exception")
        # The fields should still be set even if an exception was raised
        assert m.a == 1
        assert m.b == "test"

    @pytest.fixture(params=[{"foo": "bar"}, {"baz": 123}, {"foo": "bar", "baz": 123}])
    def context_data(self, request):
        return request.param

    def test_add(self):
        model1 = self.SimpleModel(a=1)
        model2 = self.SimpleModel(a=2)
        model = model1 + model2
        assert isinstance(model, BaseModel)
        assert model.a == 2
        assert model.b == "default"

    def test_getitem(self):
        model = self.SimpleModel(a=1)
        assert model["a"] == 1

    def test_setitem(self):
        model = self.SimpleModel(a=1)
        model["a"] = 2
        assert model.a == 2

    def test_hasattr(self):
        model = self.SimpleModel(a=1)
        assert model.hasattr("a")
        assert not model.hasattr("non_existent_key")

    def test_from_context(self, context_data):
        context = Context(context_data)
        model = self.FooModel.from_context(context)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_from_dict(self, context_data):
        model = self.FooModel.from_dict(context_data)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_from_json(self, context_data):
        json_data = json.dumps(context_data)
        model = self.FooModel.from_json(json_data)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_from_toml(self):
        toml_data = dedent(
            """
            a = 1
            b = "default"
            """
        )
        model = self.SimpleModel.from_toml(toml_data)
        assert isinstance(model, BaseModel)
        assert model.a == 1
        assert model.b == "default"

    def test_from_yaml(self, context_data):
        yaml_data = yaml.dump(context_data)
        model = self.FooModel.from_yaml(yaml_data)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_to_context(self):
        model = self.SimpleModel(a=1)
        context = model.to_context()
        assert isinstance(context, Context)
        assert context.a == 1
        assert context.b == "default"

    def test_to_dict(self):
        model = self.SimpleModel(a=1)
        dict_model = model.to_dict()
        assert isinstance(dict_model, dict)
        assert dict_model["a"] == 1
        assert dict_model["b"] == "default"

    def test_to_json(self):
        model = self.SimpleModel(a=1)
        json_model = model.to_json()
        assert isinstance(json_model, str)
        assert '"a": 1' in json_model
        assert '"b": "default"' in json_model

    def test_to_yaml(self):
        model = self.SimpleModel(a=1)
        yaml_model = model.to_yaml()
        assert isinstance(yaml_model, str)
        assert "a: 1" in yaml_model
        assert "b: default" in yaml_model

    import pytest

    class ModelWithDescription(BaseModel):
        a: int = "42"
        description: str = "This is a\nmultiline description"

    class ModelWithDocstring(BaseModel):
        """Docstring should be used as description
        when no explicit description is provided.
        """

        a: int = "42"

    class EmptyLinesShouldBeRemoved(BaseModel):
        """
        Ignore the empty line
        """

        a: int = "42"

    class ModelWithNoDescription(BaseModel):
        a: int = "42"

    class IgnoreDocstringIfDescriptionIsProvided(BaseModel):
        """This is a docstring"""

        a: int = "42"
        description: str = "This is a description"

    @pytest.mark.parametrize(
        "model_class, instance_arg, expected",
        [
            (ModelWithDescription, {"a": 1}, {"a": 1, "description": "This is a", "name": "ModelWithDescription"}),
            (
                ModelWithDocstring,
                {"a": 2},
                {"a": 2, "description": "Docstring should be used as description", "name": "ModelWithDocstring"},
            ),
            (
                EmptyLinesShouldBeRemoved,
                {"a": 3},
                {"a": 3, "description": "Ignore the empty line", "name": "EmptyLinesShouldBeRemoved"},
            ),
            (
                ModelWithNoDescription,
                {"a": 4},
                {"a": 4, "name": "ModelWithNoDescription", "description": "ModelWithNoDescription"},
            ),
            (
                IgnoreDocstringIfDescriptionIsProvided,
                {"a": 5},
                {"a": 5, "description": "This is a description", "name": "IgnoreDocstringIfDescriptionIsProvided"},
            ),
        ],
    )
    def test_name_and_multiline_description(self, model_class, instance_arg, expected):
        instance = model_class(**instance_arg)
        assert instance.model_dump() == expected

    class ModelWithLongDescription(BaseModel):
        a: int = "42"
        description: str = "This is a very long description. " * 42

    class ModelWithLongDescriptionAndNoSpaces(BaseModel):
        a: int = "42"
        description: str = "ThisIsAVeryLongDescription" * 42

    @pytest.mark.parametrize(
        "model_class, expected_length, expected_description",
        [
            (ModelWithLongDescription, 121, "This is a very long description. " * 3 + "This is a very long..."),
            (ModelWithLongDescriptionAndNoSpaces, 120, "ThisIsAVeryLongDescription" * 4 + "ThisIsAVeryLo..."),
        ],
    )
    def test_extremely_long_description(self, model_class, expected_length, expected_description):
        model = model_class()
        assert len(model.description) == expected_length
        assert model.description == expected_description
        assert model.description.endswith("...")


class TestExtraParamsMixin:
    def test_extra_params_mixin(self):
        class SimpleModelWithExtraParams(BaseModel, ExtraParamsMixin):
            a: int
            b: str = "default"

        bar = SimpleModelWithExtraParams(a=1, c=3)
        assert bar.extra_params == {"c": 3}
        assert bar.model_dump() == {
            "a": 1,
            "b": "default",
            "c": 3,
            "description": "SimpleModelWithExtraParams",
            "params": {"c": 3},
            "name": "SimpleModelWithExtraParams",
        }



"""
This module provides the RestApiReader class for interacting with RESTful APIs.

The RestApiReader class is designed to fetch data from RESTful APIs and store the response in a DataFrame. It supports
different transports, e.g. Paginated Http or Async HTTP. The main entry point is the `execute`
method, which performs transport.execute() call and provide data from the API calls.

For more details on how to use this class and its methods, refer to the class docstring.

"""

from typing import List, Tuple, Union

from pydantic import Field, InstanceOf

from pyspark.sql.types import AtomicType, StructType

from koheesio.asyncio.http import AsyncHttpGetStep
from koheesio.spark.readers import Reader
from koheesio.steps.http import HttpGetStep


class RestApiReader(Reader):
    """
    A reader class that executes an API call and stores the response in a DataFrame.

    Parameters
    ----------
    transport : Union[InstanceOf[AsyncHttpGetStep], InstanceOf[HttpGetStep]]
        The HTTP transport step.
    spark_schema : Union[str, StructType, List[str], Tuple[str, ...], AtomicType]
        The pyspark schema of the response.

    Attributes
    ----------
    transport : Union[InstanceOf[AsyncHttpGetStep], InstanceOf[HttpGetStep]]
        The HTTP transport step.
    spark_schema : Union[str, StructType, List[str], Tuple[str, ...], AtomicType]
        The pyspark schema of the response.

    Returns
    -------
    Reader.Output
        The output of the reader, which includes the DataFrame.


    Examples
    --------
    Here are some examples of how to use this class:

    Example 1: Paginated Transport
    ```python
    import requests
    from urllib3 import Retry

    from koheesio.steps.http import HttpGetStep
    from koheesio.spark.readers.rest_api import RestApiReader

    session = requests.Session()
    retry_logic = Retry(total=max_retries, status_forcelist=[503])
    session.mount("https://", HTTPAdapter(max_retries=retry_logic))
    session.mount("http://", HTTPAdapter(max_retries=retry_logic))

    transport = PaginatedHtppGetStep(
        url="https://api.example.com/data?page={page}",
        paginate=True,
        pages=3,
        session=session,
    )
    task = RestApiReader(
        transport=transport, spark_schema="id: int, page:int, value: string"
    )
    task.execute()
    all_data = [row.asDict() for row in task.output.df.collect()]
    ```

    Example 2: Async Transport
    ```python
    from aiohttp import ClientSession, TCPConnector
    from aiohttp_retry import ExponentialRetry
    from yarl import URL

    from koheesio.steps.asyncio.http import AsyncHttpGetStep
    from koheesio.spark.readers.rest_api import RestApiReader

    session = ClientSession()
    urls = [URL("http://httpbin.org/get"), URL("http://httpbin.org/get")]
    retry_options = ExponentialRetry()
    connector = TCPConnector(limit=10)
    transport = AsyncHttpGetStep(
        client_session=session,
        url=urls,
        retry_options=retry_options,
        connector=connector,
    )

    task = RestApiReader(
        transport=transport, spark_schema="id: int, page:int, value: string"
    )
    task.execute()
    all_data = [row.asDict() for row in task.output.df.collect()]
    ```

    """

    transport: Union[InstanceOf[AsyncHttpGetStep], InstanceOf[HttpGetStep]] = Field(
        ..., description="HTTP transport step", exclude=True
    )
    spark_schema: Union[str, StructType, List[str], Tuple[str, ...], AtomicType] = Field(
        ..., description="The pyspark schema of the response"
    )

    def execute(self) -> Reader.Output:
        """
        Executes the API call and stores the response in a DataFrame.

        Returns
        -------
        Reader.Output
            The output of the reader, which includes the DataFrame.
        """
        raw_data = self.transport.execute()

        if isinstance(raw_data, HttpGetStep.Output):
            data = raw_data.response_json
        elif isinstance(raw_data, AsyncHttpGetStep.Output):
            data = [d for d, _ in raw_data.responses_urls]  # type: ignore

        if data:
            self.output.df = self.spark.createDataFrame(data=data, schema=self.spark_schema)  # type: ignore



from os import environ
from unittest.mock import patch

import pytest

from pyspark.sql.types import StringType, StructField, StructType

from koheesio.spark.utils import (
    import_pandas_based_on_pyspark_version,
    on_databricks,
    schema_struct_to_schema_str,
)


def test_schema_struct_to_schema_str():
    struct_schema = StructType([StructField("a", StringType()), StructField("b", StringType())])
    val = schema_struct_to_schema_str(struct_schema)
    assert val == "a STRING,\nb STRING"
    assert schema_struct_to_schema_str(None) == ""


@pytest.mark.parametrize(
    "env_var_value, expected_result",
    [("lts_11_spark_3_scala_2.12", True), ("unit_test", True), (None, False)],
)
def test_on_databricks(env_var_value, expected_result):
    if env_var_value is not None:
        with patch.dict(environ, {"DATABRICKS_RUNTIME_VERSION": env_var_value}):
            assert on_databricks() == expected_result
    else:
        with patch.dict(environ, clear=True):
            assert on_databricks() == expected_result


@pytest.mark.parametrize(
    "spark_version, pandas_version, expected_error",
    [
        (3.3, "1.2.3", None),  # PySpark 3.3, pandas < 2, should not raise an error
        (3.4, "2.3.4", None),  # PySpark not 3.3, pandas >= 2, should not raise an error
        (3.3, "2.3.4", ImportError),  # PySpark 3.3, pandas >= 2, should raise an error
        (3.4, "1.2.3", ImportError),  # PySpark not 3.3, pandas < 2, should raise an error
    ],
)
def test_import_pandas_based_on_pyspark_version(spark_version, pandas_version, expected_error):
    with (
        patch("koheesio.spark.utils.get_spark_minor_version", return_value=spark_version),
        patch("pandas.__version__", new=pandas_version),
    ):
        if expected_error:
            with pytest.raises(expected_error):
                import_pandas_based_on_pyspark_version()
        else:
            import_pandas_based_on_pyspark_version()  # This should not raise an error



"""
Module for creating and managing Delta tables.
"""

import warnings
from typing import Dict, List, Optional, Union

from py4j.protocol import Py4JJavaError  # type: ignore

from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from koheesio.models import Field, field_validator, model_validator
from koheesio.spark import AnalysisException, SparkStep
from koheesio.spark.utils import on_databricks


class DeltaTableStep(SparkStep):
    """
    Class for creating and managing Delta tables.

    DeltaTable aims to provide a simple interface to create and manage Delta tables.
    It is a wrapper around the Spark SQL API for Delta tables.

    <!--- TODO: handle errors like this one:
    #     AnalysisException: You are trying to read a Delta table ... that does not have any columns.
    #     Write some new data with the option `mergeSchema = true` to be able to read the table.
    -->

    Example
    -------
    ```python
    from koheesio.steps import DeltaTableStep

    DeltaTableStep(
        table="my_table",
        database="my_database",
        catalog="my_catalog",
        create_if_not_exists=True,
        default_create_properties={
            "delta.randomizeFilePrefixes": "true",
            "delta.checkpoint.writeStatsAsStruct": "true",
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
        },
    )
    ```

    Methods
    -------
    get_persisted_properties()
        Get persisted properties of table.
    add_property(key: str, value: str, override: bool = False)
        Alter table and set table property.
    add_properties(properties: Dict[str, str], override: bool = False)
        Alter table and add properties.
    execute()
        Nothing to execute on a Table.
    max_version_ts_of_last_execution(query_predicate: str = None) -> datetime.datetime
        Max version timestamp of last execution. If no timestamp is found, returns 1900-01-01 00:00:00.
        Note: will raise an error if column `VERSION_TIMESTAMP` does not exist.
        <!---
        FIXME: This is not working (local or on Databricks) - asked for reference implementation through
            cop-data-engineering. Response from Sarath: New release of data common utils is coming soon in which this
            function will be fixed. He is indicating to use read_cdc instead.
        -->

    Properties
    ----------
    - name -> str
        Deprecated. Use `.table_name` instead.
    - table_name -> str
        Table name.
    - dataframe -> DataFrame
        Returns a DataFrame to be able to interact with this table.
    - columns -> Optional[List[str]]
        Returns all column names as a list.
    - has_change_type -> bool
        Checks if a column named `_change_type` is present in the table.
    - exists -> bool
        Check if table exists.

    Parameters
    ----------
    table : str
        Table name.
    database : str, optional, default=None
        Database or Schema name.
    catalog : str, optional, default=None
        Catalog name.
    create_if_not_exists : bool, optional, default=False
        Force table creation if it doesn't exist. Note: Default properties will be applied to the table during CREATION.
    default_create_properties : Dict[str, str], optional, default={"delta.randomizeFilePrefixes": "true", "delta.checkpoint.writeStatsAsStruct": "true", "delta.minReaderVersion": "2", "delta.minWriterVersion": "5"}
        Default table properties to be applied during CREATION if `force_creation` True.
    """

    # Order of inputs is reversed to ensure that TableName is always present
    table: str
    database: Optional[str] = Field(
        default=None,
        description="Database or Schema name.",
    )
    catalog: Optional[str] = Field(
        default=None,
        description="Catalog name. "
        "Note: Can be ignored if using a SparkCatalog that does not support catalog notation (e.g. Hive)",
    )

    # Table parameters
    create_if_not_exists: bool = Field(
        default=False,
        alias="force_creation",
        description="Force table creation if it doesn't exist."
        "Note: Default properties will be applied to the table during CREATION.",
    )
    default_create_properties: Dict[str, Union[str, bool, int]] = Field(
        default={
            "delta.randomizeFilePrefixes": "true",
            "delta.checkpoint.writeStatsAsStruct": "true",
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
        },
        description="Default table properties to be applied during CREATION if `create_if_not_exists` True",
    )

    @field_validator("default_create_properties")
    def _adjust_default_properties(cls, default_create_properties):
        """Adjust default properties based on environment."""
        if on_databricks():
            default_create_properties["delta.autoOptimize.autoCompact"] = True
            default_create_properties["delta.autoOptimize.optimizeWrite"] = True

        for k, v in default_create_properties.items():
            default_create_properties[k] = str(v) if not isinstance(v, bool) else str(v).lower()

        return default_create_properties

    @model_validator(mode="after")
    def _validate_catalog_database_table(self):
        """Validate that catalog, database/schema, and table are correctly set"""
        database, catalog, table = self.database, self.catalog, self.table

        try:
            self.log.debug(f"Value of `table` input parameter: {table}")
            catalog, database, table = table.split(".")
            self.log.debug("Catalog, database and table were given")
        except ValueError as e:
            if str(e) == "not enough values to unpack (expected 3, got 1)":
                self.log.debug("Only table name was given")
            elif str(e) == "not enough values to unpack (expected 3, got 2)":
                self.log.debug("Only table name and database name were given")
                database, table = table.split(".")
            else:
                raise ValueError(f"Unable to parse values for Table: {table}") from e

        self.database, self.catalog, self.table = database, catalog, table
        return self

    def get_persisted_properties(self) -> Dict[str, str]:
        """Get persisted properties of table.

        Returns
        -------
        Dict[str, str]
            Persisted properties as a dictionary.
        """
        persisted_properties = {}
        raw_options = self.spark.sql(f"SHOW TBLPROPERTIES {self.table_name}").collect()

        for ro in raw_options:
            key, value = ro.asDict().values()
            persisted_properties[key] = value

        return persisted_properties

    @property
    def is_cdf_active(self) -> bool:
        """Check if CDF property is set and activated

        Returns
        -------
        bool
            delta.enableChangeDataFeed property is set to 'true'

        """
        props = self.get_persisted_properties()
        return props.get("delta.enableChangeDataFeed", "false") == "true"

    def add_property(self, key: str, value: Union[str, int, bool], override: bool = False):
        """Alter table and set table property.

        Parameters
        ----------
        key: str
            Property key(name).
        value: Union[str, int, bool]
            Property value.
        override: bool
            Enable override of existing value for property in table.

        """
        persisted_properties = self.get_persisted_properties()
        v_str = str(value) if not isinstance(value, bool) else str(value).lower()

        def _alter_table() -> None:
            property_pair = f"'{key}'='{v_str}'"

            try:
                # noinspection SqlNoDataSourceInspection
                self.spark.sql(f"ALTER TABLE {self.table_name} SET TBLPROPERTIES ({property_pair})")
                self.log.debug(f"Table `{self.table_name}` has been altered. Property `{property_pair}` added.")
            except Py4JJavaError as e:
                msg = f"Property `{key}` can not be applied to table `{self.table_name}`. Exception: {e}"
                self.log.warning(msg)
                warnings.warn(msg)

        if self.exists:
            if key in persisted_properties and persisted_properties[key] != v_str:
                if override:
                    self.log.debug(
                        f"Property `{key}` presents in `{self.table_name}` and has value `{persisted_properties[key]}`."
                        f"Override is enabled.The value will be changed to `{v_str}`."
                    )
                    _alter_table()
                else:
                    self.log.debug(
                        f"Skipping adding property `{key}`, because it is already set "
                        f"for table `{self.table_name}` to `{v_str}`. To override it, provide override=True"
                    )
            else:
                _alter_table()
        else:
            self.default_create_properties[key] = v_str

    def add_properties(self, properties: Dict[str, Union[str, bool, int]], override: bool = False):
        """Alter table and add properties.

        Parameters
        ----------
        properties : Dict[str, Union[str, int, bool]]
            Properties to be added to table.
        override : bool, optional, default=False
            Enable override of existing value for property in table.

        """
        for k, v in properties.items():
            v_str = str(v) if not isinstance(v, bool) else str(v).lower()
            self.add_property(key=k, value=v_str, override=override)

    def execute(self):
        """Nothing to execute on a Table"""

    @property
    def table_name(self) -> str:
        """Fully qualified table name in the form of `catalog.database.table`"""
        return ".".join([n for n in [self.catalog, self.database, self.table] if n])

    @property
    def dataframe(self) -> DataFrame:
        """Returns a DataFrame to be able to interact with this table"""
        return self.spark.table(self.table_name)

    @property
    def columns(self) -> Optional[List[str]]:
        """Returns all column names as a list.

        Example
        -------
        ```python
        DeltaTableStep(...).columns
        ```
        Would for example return `['age', 'name']` if the table has columns `age` and `name`.
        """
        return self.dataframe.columns if self.exists else None

    def get_column_type(self, column: str) -> Optional[DataType]:
        """Get the type of a column in the table.

        Parameters
        ----------
        column : str
            Column name.

        Returns
        -------
        Optional[DataType]
            Column type.
        """
        return self.dataframe.schema[column].dataType if self.columns and column in self.columns else None

    @property
    def has_change_type(self) -> bool:
        """Checks if a column named `_change_type` is present in the table"""
        return "_change_type" in self.columns

    @property
    def exists(self) -> bool:
        """Check if table exists.
        Depending on the value of the boolean flag `create_if_not_exists` a different logging level is provided."""
        result = False

        try:
            self.spark.table(self.table_name)
            result = True
        except AnalysisException as e:
            err_msg = str(e).lower()
            common_message = (
                f"Table `{self.table}` doesn't exist. "
                f"The `create_if_not_exists` flag is set to {self.create_if_not_exists}."
            )

            if err_msg.startswith("[table_or_view_not_found]") or err_msg.startswith("table or view not found"):
                if self.create_if_not_exists:
                    self.log.info(" ".join((common_message, "Therefore the table will be created.")))
                else:
                    self.log.error(" ".join((common_message, "Therefore the table will not be created.")))
            else:
                raise e

        return result



import os
from unittest.mock import MagicMock, patch

import pytest
from conftest import await_job_completion
from delta import DeltaTable

from pydantic import ValidationError

from pyspark.sql import functions as F

from koheesio.spark import AnalysisException
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.writers import BatchOutputMode, StreamingOutputMode
from koheesio.spark.writers.delta import DeltaTableStreamWriter, DeltaTableWriter
from koheesio.spark.writers.delta.utils import log_clauses
from koheesio.spark.writers.stream import Trigger

pytestmark = pytest.mark.spark


def test_delta_table_writer(dummy_df, spark):
    table_name = "test_table"
    writer = DeltaTableWriter(table=table_name, output_mode=BatchOutputMode.APPEND, df=dummy_df)
    writer.execute()
    actual_count = spark.read.table(table_name).count()
    assert actual_count == 1
    writer.execute()
    actual_count = spark.read.table(table_name).count()
    assert actual_count == 2
    writer.output_mode = BatchOutputMode.OVERWRITE
    writer.execute()
    actual_count = spark.read.table(table_name).count()
    assert actual_count == 1
    writer.output_mode = BatchOutputMode.IGNORE
    writer.execute()
    actual_count = spark.read.table(table_name).count()
    assert actual_count == 1


def test_delta_partitioning(spark, sample_df_to_partition):
    table_name = "partition_table"
    DeltaTableWriter(
        table=table_name, output_mode=BatchOutputMode.OVERWRITE, df=sample_df_to_partition, partition_by=["partition"]
    ).execute()
    output_df = spark.read.table(table_name)
    assert output_df.count() == 2


def test_delta_table_merge_all(spark):
    table_name = "test_merge_all_table"
    target_df = spark.createDataFrame(
        [{"id": 1, "value": "no_merge"}, {"id": 2, "value": "expected_merge"}, {"id": 5, "value": "xxxx"}]
    )
    source_df = spark.createDataFrame(
        [
            {"id": 2, "value": "updated_value_long"},
            {"id": 3, "value": None},
            {"id": 4, "value": "new_value"},
            {"id": 5, "value": "x"},
        ]
    )
    expected = {
        1: "no_merge",
        2: "updated_value_long",
        # {"id": 3, "value": None},  No merge as new value is NONE
        4: "new_value",
        # No merge as old value is greater
        5: "xxxx",
    }
    DeltaTableWriter(table=table_name, output_mode=BatchOutputMode.APPEND, df=target_df).execute()
    DeltaTableWriter(
        table=table_name,
        output_mode=BatchOutputMode.MERGEALL,
        output_mode_params={
            "merge_cond": "target.id=source.id",
            "update_cond": F.expr("length(source.value)>length(target.value)"),
            "insert_cond": F.expr("source.value IS NOT NULL"),
        },
        df=source_df,
    ).execute()
    result = {
        list(row.asDict().values())[0]: list(row.asDict().values())[1] for row in spark.read.table(table_name).collect()
    }
    assert result == expected


def test_deltatablewriter_with_invalid_conditions(spark, dummy_df):
    table_name = "delta_test_table"
    merge_builder = (
        DeltaTable.forName(sparkSession=spark, tableOrViewName=table_name)
        .alias("target")
        .merge(condition="invalid_condition", source=dummy_df.alias("source"))
    )
    writer = DeltaTableWriter(
        table=table_name,
        output_mode=BatchOutputMode.MERGE,
        output_mode_params={"merge_builder": merge_builder},
        df=dummy_df,
    )
    with pytest.raises(AnalysisException):
        writer.execute()


@patch.dict(
    os.environ,
    {
        "DATABRICKS_RUNTIME_VERSION": "",
    },
)
def test_delta_new_table_merge(spark):
    table_name = "test_merge_new_table"
    source_df = spark.createDataFrame(
        [
            {"id": 1, "value": "new_value"},
            {"id": 2, "value": "new_value"},
            {"id": 3, "value": None},  # Will be not saved, because source.value IS NOT NULL,
            {"id": 4, "value": "new_value"},
            {"id": 5, "value": "new_value"},
        ]
    )
    expected = {1: "new_value", 2: "new_value", 4: "new_value", 5: "new_value"}
    DeltaTableWriter(
        table=DeltaTableStep(table=table_name, create_if_not_exists=True),
        output_mode=BatchOutputMode.MERGEALL,
        output_mode_params={
            "merge_cond": "target.id=source.id",
            "update_cond": "source.value>target.value",
            "insert_cond": "source.value IS NOT NULL",
        },
        df=source_df,
    ).execute()
    result = {
        list(row.asDict().values())[0]: list(row.asDict().values())[1] for row in spark.read.table(table_name).collect()
    }
    assert result == expected


@pytest.mark.parametrize(
    "class_def,bad_output_mode,streaming",
    [
        (DeltaTableWriter, "bad-value", False),
        (DeltaTableStreamWriter, "overwrite", True),
    ],
)
def test_invalid_output_mode(class_def, bad_output_mode, streaming, checkpoint_folder):
    """Passing an invalid output_mode should raise an AttributeError"""
    options = {
        "table": "table",
        "output_mode": bad_output_mode,
    }
    if streaming:
        options["checkpoint_location"] = checkpoint_folder

    with pytest.raises(ValidationError):
        class_def(**options)


@pytest.mark.parametrize(
    "constructor, output_mode, streaming, checkpoint_folder",
    [
        (DeltaTableWriter, "overwrite", False, "dummy_path"),
        (DeltaTableWriter, "append", False, "dummy_path"),
        (DeltaTableStreamWriter, "complete", True, "dummy_path"),
        (DeltaTableStreamWriter, "append", True, "dummy_path"),
    ],
)
def test_valid_output_mode(constructor, output_mode, streaming, checkpoint_folder):
    options = {
        "table": "table",
        "output_mode": output_mode,
    }
    if streaming:
        options["checkpoint_location"] = checkpoint_folder
    x = constructor(**options)
    assert x is not None


def test_delta_stream_table_writer(streaming_dummy_df, spark, checkpoint_folder):
    table_name = "test_streaming_table"
    delta_writer = DeltaTableStreamWriter(
        table=table_name,
        checkpoint_location=checkpoint_folder,
        output_mode="append",
        trigger=Trigger(processing_time="20 seconds"),
        df=streaming_dummy_df,
    )
    delta_writer.write()
    await_job_completion(timeout=20, query_id=delta_writer.streaming_query.id)
    df = spark.read.table(table_name)

    assert df.count() == 10


@pytest.mark.parametrize(
    "trigger_args,expected_exception",
    [
        ({"processing_time": "20 seconds", "available_now": True}, ValueError),
        (None, RuntimeError),
        ({"processing_time": ""}, ValueError),
        ({"continuous": "   "}, ValueError),
        ({"once": "ONCE"}, ValueError),
    ],
)
def test_delta_stream_table_writer_invalid_triggers(
    trigger_args, expected_exception, streaming_dummy_df, checkpoint_folder
):
    table_name = "test_streaming_table"
    with pytest.raises(expected_exception):
        trigger = Trigger(**trigger_args) if trigger_args else None
        DeltaTableStreamWriter(
            table=table_name,
            checkpoint_location=checkpoint_folder,
            output_mode="append",
            trigger=trigger,
            df=streaming_dummy_df,
        )


@pytest.mark.parametrize(
    "choice, options, expected",
    [
        ("APPEND", {BatchOutputMode}, BatchOutputMode.APPEND),
        ("APPEND", {StreamingOutputMode}, StreamingOutputMode.APPEND),
    ],
)
def test_get_output_mode_with_valid_choice(choice, options, expected):
    output_mode = DeltaTableWriter.get_output_mode(choice, options)
    assert output_mode == expected


@pytest.mark.parametrize(
    "choice, options",
    [
        ("INVALID", {BatchOutputMode}),
        ("INVALID", {StreamingOutputMode}),
    ],
)
def test_get_output_mode_with_invalid_choice(choice, options):
    with pytest.raises(AttributeError):
        DeltaTableWriter.get_output_mode(choice, options)


def test_delta_with_options(spark):
    """
    Checks whether the options are passed to the writer in case of batch APPEND and OVERWRITE modes
    """
    sample_df = spark.createDataFrame([{"id": 1, "value": "test_value"}])

    with patch("koheesio.spark.writers.delta.DeltaTableWriter.writer", new_callable=MagicMock) as mock_writer:
        delta_writer = DeltaTableWriter(
            table="test_table",
            output_mode=BatchOutputMode.APPEND,
            testParam1="testValue1",
            testParam2="testValue2",
            df=sample_df,
        )
        delta_writer.execute()
        mock_writer.options.assert_called_once_with(testParam1="testValue1", testParam2="testValue2")

    with patch("koheesio.spark.writers.delta.DeltaTableWriter.writer", new_callable=MagicMock) as mock_writer:
        delta_writer = DeltaTableWriter(
            table="test_table",
            output_mode=BatchOutputMode.OVERWRITE,
            testParam1="testValue1",
            testParam2="testValue2",
            df=sample_df,
        )
        delta_writer.execute()
        mock_writer.options.assert_called_once_with(testParam1="testValue1", testParam2="testValue2")


def test_merge_from_args(spark, dummy_df):
    table_name = "test_table_merge_from_args"
    dummy_df.write.format("delta").saveAsTable(table_name)

    with patch("delta.DeltaTable.merge", new_callable=MagicMock) as mock_merge:
        mock_delta_builder = MagicMock()
        mock_merge.return_value = mock_delta_builder

        # Mock the methods to return self for chaining
        mock_delta_builder.whenMatchedUpdate.return_value = mock_delta_builder
        mock_delta_builder.whenNotMatchedInsert.return_value = mock_delta_builder

        writer = DeltaTableWriter(
            df=dummy_df,
            table=table_name,
            output_mode=BatchOutputMode.MERGE,
            output_mode_params={
                "merge_builder": [
                    {"clause": "whenMatchedUpdate", "set": {"id": "source.id"}, "condition": "source.id=target.id"},
                    {
                        "clause": "whenNotMatchedInsert",
                        "values": {"id": "source.id"},
                        "condition": "source.id IS NOT NULL",
                    },
                ],
                "merge_cond": "source.id=target.id",
            },
        )
        writer._merge_builder_from_args()

        mock_delta_builder.whenMatchedUpdate.assert_called_once_with(
            set={"id": "source.id"}, condition="source.id=target.id"
        )
        mock_delta_builder.whenNotMatchedInsert.assert_called_once_with(
            values={"id": "source.id"}, condition="source.id IS NOT NULL"
        )


@pytest.mark.parametrize(
    "output_mode_params",
    [
        {
            "merge_builder": [
                {"clause": "NOT-SUPPORTED-MERGE-CLAUSE", "set": {"id": "source.id"}, "condition": "source.id=target.id"}
            ],
            "merge_cond": "source.id=target.id",
        },
        {"merge_builder": MagicMock()},
    ],
)
def test_merge_from_args_raise_value_error(spark, output_mode_params):
    with pytest.raises(ValueError):
        DeltaTableWriter(
            table="test_table_merge",
            output_mode=BatchOutputMode.MERGE,
            output_mode_params=output_mode_params,
        )


def test_merge_no_table(spark):
    table_name = "test_merge_no_table"
    target_df = spark.createDataFrame(
        [{"id": 1, "value": "no_merge"}, {"id": 2, "value": "expected_merge"}, {"id": 5, "value": "expected_merge"}]
    )
    source_df = spark.createDataFrame(
        [
            {"id": 2, "value": "longer_values_should_be_merged"},
            {"id": 3, "value": None},
            {"id": 4, "value": "not_matches_should_insert"},
            {"id": 5, "value": "longer_values_should_be_merged"},
        ]
    )
    expected = {
        1: "no_merge",
        2: "longer_values_should_be_merged",
        # 3: None - Should not be inserted, value is None
        4: "not_matches_should_insert",
        5: "longer_values_should_be_merged",
    }

    params = {
        "merge_builder": [
            {
                "clause": "whenNotMatchedInsert",
                "values": {"id": "source.id", "value": "source.value"},
                "condition": "source.value IS NOT NULL",
            },
            {
                "clause": "whenMatchedUpdate",
                "set": {"value": "source.value"},
                "condition": "length(source.value)>length(target.value)",
            },
        ],
        "merge_cond": "source.id=target.id",
    }

    DeltaTableWriter(
        df=target_df, table=table_name, output_mode=BatchOutputMode.MERGE, output_mode_params=params
    ).execute()

    DeltaTableWriter(
        df=source_df, table=table_name, output_mode=BatchOutputMode.MERGE, output_mode_params=params
    ).execute()

    result = {
        list(row.asDict().values())[0]: list(row.asDict().values())[1] for row in spark.read.table(table_name).collect()
    }

    assert result == expected


def test_log_clauses(mocker):
    # Mocking JavaObject and its methods
    mock_clauses = mocker.MagicMock()
    mock_clauses.isEmpty.return_value = False
    mock_clauses.last.return_value.nodeName.return_value = "DeltaMergeIntoTest"
    mock_clauses.length.return_value = 1

    mock_clause = mocker.MagicMock()
    mock_clause.clauseType.return_value = "test"
    mock_clause.actions.return_value.toList.return_value.apply.return_value.toString.return_value = "test_column"

    mock_condition = mocker.MagicMock()
    mock_condition.value.return_value.toString.return_value = "source_alias == target_alias"
    mock_condition.toString.return_value = "None"
    mock_clause.condition.return_value = mock_condition

    mock_clauses.apply.return_value = mock_clause

    # Call the function with the mocked JavaObject
    result = log_clauses(mock_clauses, "source_alias", "target_alias")

    # Assert the result
    assert result == "Test will perform action:Test columns (test_column) if `source_alias == target_alias`"



from unittest import mock

import pytest

from pyspark.sql import SparkSession

from koheesio.spark.readers.hana import HanaReader

pytestmark = pytest.mark.spark


class TestHanaReader:
    common_options = {
        "url": "url",
        "user": "user",
        "password": "password",
        "dbtable": "table",
    }

    def test_get_options(self):
        hana = HanaReader(**self.common_options)
        o = hana.get_options()

        assert o["driver"] == "com.sap.db.jdbc.Driver"
        assert o["fetchsize"] == 2000
        assert o["numPartitions"] == 10

    def test_execute(self, dummy_spark):
        """Method should be callable from parent class"""
        with mock.patch.object(SparkSession, "getActiveSession") as mock_spark:
            mock_spark.return_value = dummy_spark

            hana = HanaReader(**self.common_options)
            assert hana.execute().df.count() == 1



from __future__ import annotations

import io
import warnings
from copy import deepcopy
from functools import wraps
from unittest import mock
from unittest.mock import call, patch

import pytest

from pydantic import ValidationError

from pyspark.sql import DataFrame
from pyspark.sql.functions import lit

from koheesio.models import Field
from koheesio.spark.transformations.transform import Transform
from koheesio.steps import Step, StepMetaClass, StepOutput
from koheesio.steps.dummy import DummyOutput, DummyStep
from koheesio.utils import get_project_root


def dummy_function(df: DataFrame):
    return df.withColumn("hello", lit("world"))


output_dict_1 = dict(a="foo", b=42)
test_output_1 = DummyOutput(**output_dict_1)

output_dict_2 = dict(a="bar", b=69)
test_output_2 = DummyOutput(**output_dict_2)

lazy_output = DummyOutput.lazy()

# we put the newline in the description to test that the newline is removed
test_step = DummyStep(a="foo", b=2, description="Dummy step for testing purposes.\nwith a newline")
test_transformation = Transform(dummy_function)

PROJECT_ROOT = get_project_root()


class TestStepOutput:
    @pytest.mark.parametrize("output_dict, expected", [(output_dict_1, output_dict_1), (output_dict_2, output_dict_2)])
    def test_stepoutput_validate_output(self, output_dict, expected):
        """Tests that validate_output returns the expected output dict"""
        test_output = DummyOutput(**output_dict)
        actual = test_output.validate_output().model_dump()

        # we don't care about these fields
        actual.pop("name")
        actual.pop("description")

        # assert that the output is as expected
        assert actual == expected

    @pytest.mark.parametrize("output_dict", [dict(), dict(a="foo"), dict(b=42)])
    def test_stepoutput_unhappy_flow(self, output_dict):
        """Tests that validate_output raises an error when the output is not valid"""
        with pytest.raises(ValidationError):
            DummyOutput(**output_dict)

    @pytest.mark.parametrize(
        "a, b, expected",
        [
            # unhappy flow
            (None, None, ValidationError),
            # happy flow
            (
                "foo",
                42,
                {"a": "foo", "b": 42, "name": "DummyOutput", "description": "Dummy output for testing purposes."},
            ),
            # test wrong type assigned
            ("foo", "invalid type", ValidationError),
        ],
    )
    def test_stepoutput_lazy(self, a, b, expected):
        """Tests that lazy output works as expected, also tests validate_output"""
        lazy_output = DummyOutput.lazy()
        lazy_output.a = a
        lazy_output.b = b

        with warnings.catch_warnings(record=True):
            warnings.filterwarnings(
                "ignore",
                message=r"^'Pydantic serializer warnings:\n  Expected `int` but got `str` - serialized value may not be as expected'",
                category=UserWarning,
            )
            if expected == ValidationError:
                with pytest.raises(expected):
                    lazy_output.validate_output()
            else:
                actual = lazy_output.validate_output().model_dump()
                print(f"{actual=}")
                assert actual == expected

    @pytest.mark.parametrize("attribute, expected", [("a", True), ("d", False)])
    def test_stepoutput_hasattr(self, attribute, expected):
        """Tests that hasattr works as expected"""
        assert hasattr(test_output_1, attribute) == expected

    def test_stepoutput_getters_and_setters(self):
        """Tests that getters and setters work as expected"""
        test_value = "bar"

        # __setitem__ and __getitem__
        test_a = deepcopy(test_output_1)
        test_a.a = test_value
        assert test_a.a == test_value

        # test subscriptable getter and setter
        test_b = deepcopy(test_output_1)
        test_b["a"] = test_value
        assert test_b["a"] == test_value

        # test get and set methods
        test_c = deepcopy(test_output_1)
        test_c.set("a", test_value)
        assert test_c.get("a") == test_value

        # test get with default value
        test_d = deepcopy(test_output_1)
        expected = "default is working"
        actual = test_d.get("non_existent", expected)
        assert actual == expected

    def test_stepoutput_merge_and_add(self):
        """Tests that merge and add work as expected"""
        expected = {**output_dict_1, **output_dict_2}

        # test merge
        test_a = deepcopy(test_output_1)
        test_a.merge(test_output_2)
        actual = test_a.model_dump()
        actual.pop("name")
        actual.pop("description")
        assert actual == expected

        # test add (through += notation)
        test_b = deepcopy(test_output_1)
        test_b += test_output_2
        actual = test_b.model_dump()
        actual.pop("name")
        actual.pop("description")
        assert actual == expected

        # test add (through + notation)
        test_c = deepcopy(test_output_1)
        test_c + test_output_2
        actual = test_c.model_dump()
        actual.pop("name")
        actual.pop("description")
        assert actual == expected


class TestStep:
    def test_simple_step(self):
        class SimpleStep(Step):
            a: str

            class Output(StepOutput):
                b: str

            def execute(self) -> Output:
                self.output.b = f"{self.a}-some-suffix"

        step = SimpleStep(a="foo")

        assert step.model_dump() == dict(a="foo", description="SimpleStep", name="SimpleStep")
        assert step.execute().model_dump() == dict(
            b="foo-some-suffix", name="SimpleStep.Output", description="Output for SimpleStep"
        )

        # as long as the following doesn't raise an error, we're good
        lazy_step_output = SimpleStep.Output.lazy()
        lazy_step_output.b = "bar"
        lazy_step_output.validate_output()

    def test_step_execute_and_run(self):
        """test that execute returns implicitly and that output is accessible
        note: this ensures that the metaclass wrapper is working as expected
        """
        actual_execute = test_step.execute()
        actual_run = test_step.run()
        expected = test_step.Output(
            name=test_step.name + ".Output",
            description=f"Output for {test_step.name}",
            a="foo",
            b=2,
            c="foofoo",
        )
        print(f"{actual_execute=}")
        print(f"{actual_run=}")
        print(f"{expected=}")
        assert actual_execute == actual_run == expected

        # 3 ways to retrieve output
        actual = actual_execute
        assert actual.get("a") == actual.a == actual["a"] == expected["a"]
        assert actual.get("b") == actual.b == actual["b"] == expected["b"]
        assert actual.get("c") == actual.c == actual["c"] == expected["c"]

    def test_step_to_yaml_description_and_name(self):
        """Tests that to_yaml works as expected"""
        # test that description and name are set correctly prior to executing
        assert test_step.name == "DummyStep"
        assert test_step.description == "Dummy step for testing purposes."

        # execute the step
        test_step.execute()
        expected_output_path = PROJECT_ROOT / "tests" / "_data" / "steps"

        # using .strip() to avoid the test failing on leading or trailing whitespaces
        expected = (expected_output_path / "expected_step_output.yaml").read_text().strip()
        actual = test_step.repr_yaml().strip()
        assert actual == expected

        # testing with simple=True
        expected_simple = (expected_output_path / "expected_step_output_simple.yaml").read_text().strip()
        actual_simple = test_step.repr_yaml(simple=True).strip()
        assert actual != actual_simple == expected_simple

    def test_get_description_and_name(self):
        assert test_step.name == "DummyStep"
        assert test_step.description == "Dummy step for testing purposes."

    def test_repr_str_dunder(self):
        # using .strip() to avoid the test failing on leading or trailing whitespaces
        actual = str(DummyStep(a="foo", b=2)).strip()
        expected = "DummyStep\n" "=========\n" "input:\n" "  a: foo\n" "  b: 2"
        assert actual == expected

    def test_step_with_super_call(self):
        class MyCustomParentStep(Step):
            foo: str = Field(default=..., description="Foo")
            bar: str = Field(default=..., description="Bar")

            class Output(Step.Output):
                baz: str = Field(default=..., description="Baz")

            def execute(self) -> Output:
                self.output.baz = self.foo + self.bar

        class MyCustomChildStepSuperCallFirst(MyCustomParentStep):
            class Output(MyCustomParentStep.Output):
                qux: str = Field(default=..., description="Qux")

            def execute(self) -> Output:
                super().execute()
                self.output.qux = self.foo + "qux"

        class MyCustomChildStepSuperCallAfter(MyCustomParentStep):
            class Output(MyCustomParentStep.Output):
                qux: str = Field(default=..., description="Qux")

            def execute(self) -> Output:
                self.output.qux = self.foo + "qux"
                super().execute()

        super_call_after_output = MyCustomChildStepSuperCallAfter(foo="foo", bar="bar", qux="qux")
        assert super_call_after_output.execute() is not None

        super_call_first_output = MyCustomChildStepSuperCallFirst(foo="foo", bar="bar", qux="qux")
        assert super_call_first_output.execute() is not None

    def test_execute_wrapper_only_called_once_when_nested(self):
        """
        Tests that _execute_wrapper is only called once when nested multiple times

        This can happen when a step is inherited from another step, and the child step has a _execute_wrapper decorator
        applied to it. In this case, the _execute_wrapper decorator could be called multiple times, which is not what we want.
        This test checks that Koheesio is not calling _execute_wrapper multiple times.
        """

        class FooStep(Step):
            foo: str

            class Output(Step.Output):
                bar: str

            @wraps(StepMetaClass._execute_wrapper)
            @wraps(StepMetaClass._execute_wrapper)
            @wraps(StepMetaClass._execute_wrapper)
            @wraps(StepMetaClass._execute_wrapper)
            @wraps(StepMetaClass._execute_wrapper)
            @wraps(StepMetaClass._execute_wrapper)
            def execute(self):
                self.output.bar = self.foo

        obj = FooStep(foo="bar")
        obj.execute()

        # Check that do_execute was not called multiple times
        assert (
            getattr(getattr(obj.execute, "_partialmethod", None), "_step_execute_wrapper_sentinel", None)
            is StepMetaClass._step_execute_wrapper_sentinel
        )
        assert getattr(getattr(obj.execute, "_partialmethod", None), "_wrap_count", 0) == 1

    class YourClass(Step):
        def execute(self):
            self.log.info(f"This is from the execute method of {self.__class__.__name__}")

    class YourClass2(YourClass):
        def execute(self):
            self.log.info(f"This is from the execute method of {self.__class__.__name__}")

    class YourClass3(YourClass2):
        def execute(self):
            self.log.info(f"This is from the execute method of {self.__class__.__name__}")

    class MyMetaClass(StepMetaClass):
        @classmethod
        def _log_end_message(cls, step: Step, skip_logging: bool = False, *args, **kwargs):
            print("It's me from custom meta class")
            super()._log_end_message(step, skip_logging, *args, **kwargs)

    class MyMetaClass2(StepMetaClass):
        @classmethod
        def _validate_output(cls, step: Step, skip_validating: bool = False, *args, **kwargs):
            step.output.dummy_value = "dummy"

    class YourClassWithCustomMeta(Step, metaclass=MyMetaClass):
        def execute(self):
            self.log.info(f"This is from the execute method of {self.__class__.__name__}")

    class YourClassWithCustomMeta2(Step, metaclass=MyMetaClass2):
        def execute(self):
            self.log.info(f"This is from the execute method of {self.__class__.__name__}")

    @pytest.mark.parametrize("test_class", [YourClassWithCustomMeta])
    def test_custom_metaclass_log(self, test_class):
        with (
            patch.object(test_class, "log", autospec=True) as mock_log,
            patch("sys.stdout", new=io.StringIO()) as mock_stdout,
        ):
            obj = test_class()
            obj.execute()

            # Get the output of print statements
            print_output = mock_stdout.getvalue()

            name = test_class.__name__

            # Check that logs were called once (and only once) with the correct messages, and in the correct order
            calls = [
                call.info("Start running step"),
                call.debug(f"Step Input: name='{name}' description='{name}'"),
                call.info(f"This is from the execute method of {name}"),
                call.debug(f"Step Output: name='{name}.Output' description='Output for {name}'"),
                call.info("Finished running step"),
            ]
            mock_log.assert_has_calls(calls, any_order=False)

            assert "It's me from custom meta class" in print_output

    @pytest.mark.parametrize("test_class", [YourClassWithCustomMeta2])
    def test_custom_metaclass_output(self, test_class):
        with patch.object(test_class, "log", autospec=True) as mock_log:
            obj = test_class()
            obj.execute()

            name = test_class.__name__

            # Check that logs were called once (and only once) with the correct messages, and in the correct order
            calls = [
                call.info("Start running step"),
                call.debug(f"Step Input: name='{name}' description='{name}'"),
                call.info(f"This is from the execute method of {name}"),
                call.debug(f"Step Output: name='{name}.Output' description='Output for {name}' dummy_value='dummy'"),
                call.info("Finished running step"),
            ]
            mock_log.assert_has_calls(calls, any_order=False)

            assert obj.output.dummy_value == "dummy"

    @pytest.mark.parametrize("test_class", [YourClass, YourClass2, YourClass3])
    def test_log_called_once(self, test_class):
        with mock.patch.object(test_class, "log", autospec=True) as mock_log:
            obj = test_class()
            obj.execute()

            name = test_class.__name__

            # Check that logs were called once (and only once) with the correct messages, and in the correct order
            calls = [
                call.info("Start running step"),
                call.debug(f"Step Input: name='{name}' description='{name}'"),
                call.info(f"This is from the execute method of {name}"),
                call.debug(f"Step Output: name='{name}.Output' description='Output for {name}'"),
                call.info("Finished running step"),
            ]
            mock_log.assert_has_calls(calls, any_order=False)

    def test_output_validation(self):
        class MyStep(Step):
            a: str
            b: int

            class Output(Step.Output):
                c: str
                d: int

            def execute(self):
                self.output.c = self.a
                self.output.d = self.b

        with mock.patch.object(Step.Output, "validate_output", autospec=True) as mock_validator:
            step = MyStep(a="foo", b=42)
            step.execute()

            mock_validator.assert_called_once()



"""
Snowflake steps and tasks for Koheesio

Every class in this module is a subclass of `Step` or `Task` and is used to perform operations on Snowflake.

Notes
-----
Every Step in this module is based on [SnowflakeBaseModel](./snowflake.md#koheesio.spark.snowflake.SnowflakeBaseModel).
The following parameters are available for every Step.

Parameters
----------
url : str
    Hostname for the Snowflake account, e.g. <account>.snowflakecomputing.com.
    Alias for `sfURL`.
user : str
    Login name for the Snowflake user.
    Alias for `sfUser`.
password : SecretStr
    Password for the Snowflake user.
    Alias for `sfPassword`.
database : str
    The database to use for the session after connecting.
    Alias for `sfDatabase`.
sfSchema : str
    The schema to use for the session after connecting.
    Alias for `schema` ("schema" is a reserved name in Pydantic, so we use `sfSchema` as main name instead).
role : str
    The default security role to use for the session after connecting.
    Alias for `sfRole`.
warehouse : str
    The default virtual warehouse to use for the session after connecting.
    Alias for `sfWarehouse`.
authenticator : Optional[str], optional, default=None
    Authenticator for the Snowflake user. Example: "okta.com".
options : Optional[Dict[str, Any]], optional, default={"sfCompress": "on", "continue_on_error": "off"}
    Extra options to pass to the Snowflake connector.
format : str, optional, default="snowflake"
    The default `snowflake` format can be used natively in Databricks, use `net.snowflake.spark.snowflake` in other
    environments and make sure to install required JARs.
"""

import json
from typing import Any, Dict, List, Optional, Set, Union
from abc import ABC
from copy import deepcopy
from textwrap import dedent

from pyspark.sql import DataFrame, Window
from pyspark.sql import functions as f
from pyspark.sql import types as t

from koheesio import Step, StepOutput
from koheesio.logger import LoggingFactory, warn
from koheesio.models import (
    BaseModel,
    ExtraParamsMixin,
    Field,
    SecretStr,
    conlist,
    field_validator,
    model_validator,
)
from koheesio.spark import SparkStep
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.readers.delta import DeltaTableReader, DeltaTableStreamReader
from koheesio.spark.readers.jdbc import JdbcReader
from koheesio.spark.transformations import Transformation
from koheesio.spark.writers import BatchOutputMode, Writer
from koheesio.spark.writers.stream import (
    ForEachBatchStreamWriter,
    writer_to_foreachbatch,
)

__all__ = [
    "AddColumn",
    "CreateOrReplaceTableFromDataFrame",
    "DbTableQuery",
    "GetTableSchema",
    "GrantPrivilegesOnFullyQualifiedObject",
    "GrantPrivilegesOnObject",
    "GrantPrivilegesOnTable",
    "GrantPrivilegesOnView",
    "Query",
    "RunQuery",
    "SnowflakeBaseModel",
    "SnowflakeReader",
    "SnowflakeStep",
    "SnowflakeTableStep",
    "SnowflakeTransformation",
    "SnowflakeWriter",
    "SyncTableAndDataFrameSchema",
    "SynchronizeDeltaToSnowflakeTask",
    "TableExists",
]

# pylint: disable=inconsistent-mro, too-many-lines
# Turning off inconsistent-mro because we are using ABCs and Pydantic models and Tasks together in the same class
# Turning off too-many-lines because we are defining a lot of classes in this file


class SnowflakeBaseModel(BaseModel, ExtraParamsMixin, ABC):
    """
    BaseModel for setting up Snowflake Driver options.

    Notes
    -----
    * Snowflake is supported natively in Databricks 4.2 and newer:
        https://docs.snowflake.com/en/user-guide/spark-connector-databricks
    * Refer to Snowflake docs for the installation instructions for non-Databricks environments:
        https://docs.snowflake.com/en/user-guide/spark-connector-install
    * Refer to Snowflake docs for connection options:
        https://docs.snowflake.com/en/user-guide/spark-connector-use#setting-configuration-options-for-the-connector

    Parameters
    ----------
    url : str
        Hostname for the Snowflake account, e.g. <account>.snowflakecomputing.com.
        Alias for `sfURL`.
    user : str
        Login name for the Snowflake user.
        Alias for `sfUser`.
    password : SecretStr
        Password for the Snowflake user.
        Alias for `sfPassword`.
    database : str
        The database to use for the session after connecting.
        Alias for `sfDatabase`.
    sfSchema : str
        The schema to use for the session after connecting.
        Alias for `schema` ("schema" is a reserved name in Pydantic, so we use `sfSchema` as main name instead).
    role : str
        The default security role to use for the session after connecting.
        Alias for `sfRole`.
    warehouse : str
        The default virtual warehouse to use for the session after connecting.
        Alias for `sfWarehouse`.
    authenticator : Optional[str], optional, default=None
        Authenticator for the Snowflake user. Example: "okta.com".
    options : Optional[Dict[str, Any]], optional, default={"sfCompress": "on", "continue_on_error": "off"}
        Extra options to pass to the Snowflake connector.
    format : str, optional, default="snowflake"
        The default `snowflake` format can be used natively in Databricks, use `net.snowflake.spark.snowflake` in other
        environments and make sure to install required JARs.

    """

    url: str = Field(
        default=...,
        alias="sfURL",
        description="Hostname for the Snowflake account, e.g. <account>.snowflakecomputing.com",
        examples=["example.snowflakecomputing.com"],
    )
    user: str = Field(default=..., alias="sfUser", description="Login name for the Snowflake user")
    password: SecretStr = Field(default=..., alias="sfPassword", description="Password for the Snowflake user")
    authenticator: Optional[str] = Field(
        default=None,
        description="Authenticator for the Snowflake user",
        examples=["okta.com"],
    )
    database: str = Field(
        default=..., alias="sfDatabase", description="The database to use for the session after connecting"
    )
    sfSchema: str = Field(default=..., alias="schema", description="The schema to use for the session after connecting")
    role: str = Field(
        default=..., alias="sfRole", description="The default security role to use for the session after connecting"
    )
    warehouse: str = Field(
        default=...,
        alias="sfWarehouse",
        description="The default virtual warehouse to use for the session after connecting",
    )
    options: Optional[Dict[str, Any]] = Field(
        default={"sfCompress": "on", "continue_on_error": "off"},
        description="Extra options to pass to the Snowflake connector",
    )
    format: str = Field(
        default="snowflake",
        description="The default `snowflake` format can be used natively in Databricks, use "
        "`net.snowflake.spark.snowflake` in other environments and make sure to install required JARs.",
    )

    def get_options(self):
        """Get the sfOptions as a dictionary."""
        return {
            key: value
            for key, value in {
                "sfURL": self.url,
                "sfUser": self.user,
                "sfPassword": self.password.get_secret_value(),
                "authenticator": self.authenticator,
                "sfDatabase": self.database,
                "sfSchema": self.sfSchema,
                "sfRole": self.role,
                "sfWarehouse": self.warehouse,
                **self.options,
            }.items()
            if value is not None
        }


class SnowflakeStep(SnowflakeBaseModel, SparkStep, ABC):
    """Expands the SnowflakeBaseModel so that it can be used as a Step"""


class SnowflakeTableStep(SnowflakeStep, ABC):
    """Expands the SnowflakeStep, adding a 'table' parameter"""

    table: str = Field(default=..., description="The name of the table")

    def get_options(self):
        options = super().get_options()
        options["table"] = self.table
        return options


class SnowflakeReader(SnowflakeBaseModel, JdbcReader):
    """
    Wrapper around JdbcReader for Snowflake.

    Example
    -------
    ```python
    sr = SnowflakeReader(
        url="foo.snowflakecomputing.com",
        user="YOUR_USERNAME",
        password="***",
        database="db",
        schema="schema",
    )
    df = sr.read()
    ```

    Notes
    -----
    * Snowflake is supported natively in Databricks 4.2 and newer:
        https://docs.snowflake.com/en/user-guide/spark-connector-databricks
    * Refer to Snowflake docs for the installation instructions for non-Databricks environments:
        https://docs.snowflake.com/en/user-guide/spark-connector-install
    * Refer to Snowflake docs for connection options:
        https://docs.snowflake.com/en/user-guide/spark-connector-use#setting-configuration-options-for-the-connector
    """

    driver: Optional[str] = None  # overriding `driver` property of JdbcReader, because it is not required by Snowflake


class SnowflakeTransformation(SnowflakeBaseModel, Transformation, ABC):
    """Adds Snowflake parameters to the Transformation class"""


class RunQuery(SnowflakeStep):
    """
    Run a query on Snowflake that does not return a result, e.g. create table statement

    This is a wrapper around 'net.snowflake.spark.snowflake.Utils.runQuery' on the JVM

    Example
    -------
    ```python
    RunQuery(
        database="MY_DB",
        schema="MY_SCHEMA",
        warehouse="MY_WH",
        user="account",
        password="***",
        role="APPLICATION.SNOWFLAKE.ADMIN",
        query="CREATE TABLE test (col1 string)",
    ).execute()
    ```
    """

    query: str = Field(default=..., description="The query to run", alias="sql")

    @field_validator("query")
    def validate_query(cls, query):
        """Replace escape characters"""
        return query.replace("\\n", "\n").replace("\\t", "\t").strip()

    def get_options(self):
        # Executing the RunQuery without `host` option in Databricks throws:
        # An error occurred while calling z:net.snowflake.spark.snowflake.Utils.runQuery.
        # : java.util.NoSuchElementException: key not found: host
        options = super().get_options()
        options["host"] = options["sfURL"]
        return options

    def execute(self) -> None:
        if not self.query:
            self.log.warning("Empty string given as query input, skipping execution")
            return
        # noinspection PyProtectedMember
        self.spark._jvm.net.snowflake.spark.snowflake.Utils.runQuery(self.get_options(), self.query)


class Query(SnowflakeReader):
    """
    Query data from Snowflake and return the result as a DataFrame

    Example
    -------
    ```python
    Query(
        database="MY_DB",
        schema_="MY_SCHEMA",
        warehouse="MY_WH",
        user="gid.account@nike.com",
        password=Secret("super-secret-password"),
        role="APPLICATION.SNOWFLAKE.ADMIN",
        query="SELECT * FROM MY_TABLE",
    ).execute().df
    ```
    """

    query: str = Field(default=..., description="The query to run")

    @field_validator("query")
    def validate_query(cls, query):
        """Replace escape characters"""
        query = query.replace("\\n", "\n").replace("\\t", "\t").strip()
        return query

    def get_options(self):
        """add query to options"""
        options = super().get_options()
        options["query"] = self.query
        return options


class DbTableQuery(SnowflakeReader):
    """
    Read table from Snowflake using the `dbtable` option instead of `query`

    Example
    -------
    ```python
    DbTableQuery(
        database="MY_DB",
        schema_="MY_SCHEMA",
        warehouse="MY_WH",
        user="user",
        password=Secret("super-secret-password"),
        role="APPLICATION.SNOWFLAKE.ADMIN",
        table="db.schema.table",
    ).execute().df
    ```
    """

    dbtable: str = Field(default=..., alias="table", description="The name of the table")


class TableExists(SnowflakeTableStep):
    """
    Check if the table exists in Snowflake by using INFORMATION_SCHEMA.

    Example
    -------
    ```python
    k = TableExists(
        url="foo.snowflakecomputing.com",
        user="YOUR_USERNAME",
        password="***",
        database="db",
        schema="schema",
        table="table",
    )
    ```
    """

    class Output(StepOutput):
        """Output class for TableExists"""

        exists: bool = Field(default=..., description="Whether or not the table exists")

    def execute(self):
        query = (
            dedent(
                # Force upper case, due to case-sensitivity of where clause
                f"""
            SELECT *
            FROM INFORMATION_SCHEMA.TABLES
            WHERE TABLE_CATALOG     = '{self.database}'
              AND TABLE_SCHEMA      = '{self.sfSchema}'
              AND TABLE_TYPE        = 'BASE TABLE'
              AND upper(TABLE_NAME) = '{self.table.upper()}'
            """  # nosec B608: hardcoded_sql_expressions
            )
            .upper()
            .strip()
        )

        self.log.debug(f"Query that was executed to check if the table exists:\n{query}")

        df = Query(**self.get_options(), query=query).read()

        exists = df.count() > 0
        self.log.info(f"Table {self.table} {'exists' if exists else 'does not exist'}")
        self.output.exists = exists


def map_spark_type(spark_type: t.DataType):
    """
    Translates Spark DataFrame Schema type to SnowFlake type

    | Basic Types       | Snowflake Type |
    |-------------------|----------------|
    | StringType        | STRING         |
    | NullType          | STRING         |
    | BooleanType       | BOOLEAN        |

    | Numeric Types     | Snowflake Type |
    |-------------------|----------------|
    | LongType          | BIGINT         |
    | IntegerType       | INT            |
    | ShortType         | SMALLINT       |
    | DoubleType        | DOUBLE         |
    | FloatType         | FLOAT          |
    | NumericType       | FLOAT          |
    | ByteType          | BINARY         |

    | Date / Time Types | Snowflake Type |
    |-------------------|----------------|
    | DateType          | DATE           |
    | TimestampType     | TIMESTAMP      |

    | Advanced Types    | Snowflake Type |
    |-------------------|----------------|
    | DecimalType       | DECIMAL        |
    | MapType           | VARIANT        |
    | ArrayType         | VARIANT        |
    | StructType        | VARIANT        |

    References
    ----------
    - Spark SQL DataTypes: https://spark.apache.org/docs/latest/sql-ref-datatypes.html
    - Snowflake DataTypes: https://docs.snowflake.com/en/sql-reference/data-types.html

    Parameters
    ----------
    spark_type : pyspark.sql.types.DataType
        DataType taken out of the StructField

    Returns
    -------
    str
        The Snowflake data type
    """
    # StructField means that the entire Field was passed, we need to extract just the dataType before continuing
    if isinstance(spark_type, t.StructField):
        spark_type = spark_type.dataType

    # Check if the type is DayTimeIntervalType
    if isinstance(spark_type, t.DayTimeIntervalType):
        warn(
            "DayTimeIntervalType is being converted to STRING. "
            "Consider converting to a more supported date/time/timestamp type in Snowflake."
        )

    # fmt: off
    # noinspection PyUnresolvedReferences
    data_type_map = {
        # Basic Types
        t.StringType: "STRING",
        t.NullType: "STRING",
        t.BooleanType: "BOOLEAN",

        # Numeric Types
        t.LongType: "BIGINT",
        t.IntegerType: "INT",
        t.ShortType: "SMALLINT",
        t.DoubleType: "DOUBLE",
        t.FloatType: "FLOAT",
        t.NumericType: "FLOAT",
        t.ByteType: "BINARY",
        t.BinaryType: "VARBINARY",

        # Date / Time Types
        t.DateType: "DATE",
        t.TimestampType: "TIMESTAMP",
        t.DayTimeIntervalType: "STRING",

        # Advanced Types
        t.DecimalType:
            f"DECIMAL({spark_type.precision},{spark_type.scale})"  # pylint: disable=no-member
            if isinstance(spark_type, t.DecimalType) else "DECIMAL(38,0)",
        t.MapType: "VARIANT",
        t.ArrayType: "VARIANT",
        t.StructType: "VARIANT",
    }
    return data_type_map.get(type(spark_type), 'STRING')
    # fmt: on


class CreateOrReplaceTableFromDataFrame(SnowflakeTransformation):
    """
    Create (or Replace) a Snowflake table which has the same schema as a Spark DataFrame

    Can be used as any Transformation. The DataFrame is however left unchanged, and only used for determining the
    schema of the Snowflake Table that is to be created (or replaced).

    Example
    -------
    ```python
    CreateOrReplaceTableFromDataFrame(
        database="MY_DB",
        schema="MY_SCHEMA",
        warehouse="MY_WH",
        user="gid.account@nike.com",
        password="super-secret-password",
        role="APPLICATION.SNOWFLAKE.ADMIN",
        table="MY_TABLE",
        df=df,
    ).execute()
    ```

    Or, as a Transformation:
    ```python
    CreateOrReplaceTableFromDataFrame(
        ...
        table="MY_TABLE",
    ).transform(df)
    ```

    """

    table: str = Field(default=..., alias="table_name", description="The name of the (new) table")

    class Output(SnowflakeTransformation.Output):
        """Output class for CreateOrReplaceTableFromDataFrame"""

        input_schema: t.StructType = Field(default=..., description="The original schema from the input DataFrame")
        snowflake_schema: str = Field(
            default=..., description="Derived Snowflake table schema based on the input DataFrame"
        )
        query: str = Field(default=..., description="Query that was executed to create the table")

    def execute(self):
        self.output.df = self.df

        input_schema = self.df.schema
        self.output.input_schema = input_schema

        snowflake_schema = ", ".join([f"{c.name} {map_spark_type(c.dataType)}" for c in input_schema])
        self.output.snowflake_schema = snowflake_schema

        table_name = f"{self.database}.{self.sfSchema}.{self.table}"
        query = f"CREATE OR REPLACE TABLE {table_name} ({snowflake_schema})"
        self.output.query = query

        RunQuery(**self.get_options(), query=query).execute()


class GrantPrivilegesOnObject(SnowflakeStep):
    """
    A wrapper on Snowflake GRANT privileges

    With this Step, you can grant Snowflake privileges to a set of roles on a table, a view, or an object

    See Also
    --------
    https://docs.snowflake.com/en/sql-reference/sql/grant-privilege.html

    Parameters
    ----------
    warehouse : str
        The name of the warehouse. Alias for `sfWarehouse`
    user : str
        The username. Alias for `sfUser`
    password : SecretStr
        The password. Alias for `sfPassword`
    role : str
        The role name
    object : str
        The name of the object to grant privileges on
    type : str
        The type of object to grant privileges on, e.g. TABLE, VIEW
    privileges : Union[conlist(str, min_length=1), str]
        The Privilege/Permission or list of Privileges/Permissions to grant on the given object.
    roles : Union[conlist(str, min_length=1), str]
        The Role or list of Roles to grant the privileges to

    Example
    -------
    ```python
    GrantPermissionsOnTable(
        object="MY_TABLE",
        type="TABLE",
        warehouse="MY_WH",
        user="gid.account@nike.com",
        password=Secret("super-secret-password"),
        role="APPLICATION.SNOWFLAKE.ADMIN",
        permissions=["SELECT", "INSERT"],
    ).execute()
    ```

    In this example, the `APPLICATION.SNOWFLAKE.ADMIN` role will be granted `SELECT` and `INSERT` privileges on
    the `MY_TABLE` table using the `MY_WH` warehouse.
    """

    object: str = Field(default=..., description="The name of the object to grant privileges on")
    type: str = Field(default=..., description="The type of object to grant privileges on, e.g. TABLE, VIEW")

    privileges: Union[conlist(str, min_length=1), str] = Field(
        default=...,
        alias="permissions",
        description="The Privilege/Permission or list of Privileges/Permissions to grant on the given object. "
        "See https://docs.snowflake.com/en/sql-reference/sql/grant-privilege.html",
    )
    roles: Union[conlist(str, min_length=1), str] = Field(
        default=...,
        alias="role",
        validation_alias="roles",
        description="The Role or list of Roles to grant the privileges to",
    )

    class Output(SnowflakeStep.Output):
        """Output class for GrantPrivilegesOnObject"""

        query: conlist(str, min_length=1) = Field(
            default=..., description="Query that was executed to grant privileges", validate_default=False
        )

    @model_validator(mode="before")
    def set_roles_privileges(cls, values):
        """Coerce roles and privileges to be lists if they are not already."""
        roles_value = values.get("roles") or values.get("role")
        privileges_value = values.get("privileges")

        if not (roles_value and privileges_value):
            raise ValueError("You have to specify roles AND privileges when using 'GrantPrivilegesOnObject'.")

        # coerce values to be lists
        values["roles"] = [roles_value] if isinstance(roles_value, str) else roles_value
        values["role"] = values["roles"][0]  # hack to keep the validator happy
        values["privileges"] = [privileges_value] if isinstance(privileges_value, str) else privileges_value

        return values

    @model_validator(mode="after")
    def validate_object_and_object_type(self):
        """Validate that the object and type are set."""
        object_value = self.object
        if not object_value:
            raise ValueError("You must provide an `object`, this should be the name of the object. ")

        object_type = self.type
        if not object_type:
            raise ValueError(
                "You must provide a `type`, e.g. TABLE, VIEW, DATABASE. "
                "See https://docs.snowflake.com/en/sql-reference/sql/grant-privilege.html"
            )

        return self

    def get_query(self, role: str):
        """Build the GRANT query

        Parameters
        ----------
        role: str
            The role name

        Returns
        -------
        query : str
            The Query that performs the grant
        """
        query = f"GRANT {','.join(self.privileges)} ON {self.type} {self.object} TO ROLE {role}".upper()
        return query

    def execute(self):
        self.output.query = []
        roles = self.roles

        for role in roles:
            query = self.get_query(role)
            self.output.query.append(query)
            RunQuery(**self.get_options(), query=query).execute()


class GrantPrivilegesOnFullyQualifiedObject(GrantPrivilegesOnObject):
    """Grant Snowflake privileges to a set of roles on a fully qualified object, i.e. `database.schema.object_name`

    This class is a subclass of `GrantPrivilegesOnObject` and is used to grant privileges on a fully qualified object.
    The advantage of using this class is that it sets the object name to be fully qualified, i.e.
    `database.schema.object_name`.

    Meaning, you can set the `database`, `schema` and `object` separately and the object name will be set to be fully
    qualified, i.e. `database.schema.object_name`.

    Example
    -------
    ```python
    GrantPrivilegesOnFullyQualifiedObject(
        database="MY_DB",
        schema="MY_SCHEMA",
        warehouse="MY_WH",
        ...
        object="MY_TABLE",
        type="TABLE",
        ...
    )
    ```

    In this example, the object name will be set to be fully qualified, i.e. `MY_DB.MY_SCHEMA.MY_TABLE`.
    If you were to use `GrantPrivilegesOnObject` instead, you would have to set the object name to be fully qualified
    yourself.
    """

    @model_validator(mode="after")
    def set_object_name(self):
        """Set the object name to be fully qualified, i.e. database.schema.object_name"""
        # database, schema, obj_name
        db = self.database
        schema = self.model_dump()["sfSchema"]  # since "schema" is a reserved name
        obj_name = self.object

        self.object = f"{db}.{schema}.{obj_name}"

        return self


class GrantPrivilegesOnTable(GrantPrivilegesOnFullyQualifiedObject):
    """Grant Snowflake privileges to a set of roles on a table"""

    type: str = "TABLE"
    object: str = Field(
        default=...,
        alias="table",
        description="The name of the Table to grant Privileges on. This should be just the name of the table; so "
        "without Database and Schema, use sfDatabase/database and sfSchema/schema to set those instead.",
    )


class GrantPrivilegesOnView(GrantPrivilegesOnFullyQualifiedObject):
    """Grant Snowflake privileges to a set of roles on a view"""

    type: str = "VIEW"
    object: str = Field(
        default=...,
        alias="view",
        description="The name of the View to grant Privileges on. This should be just the name of the view; so "
        "without Database and Schema, use sfDatabase/database and sfSchema/schema to set those instead.",
    )


class GetTableSchema(SnowflakeStep):
    """
    Get the schema from a Snowflake table as a Spark Schema

    Notes
    -----
    * This Step will execute a `SELECT * FROM <table> LIMIT 1` query to get the schema of the table.
    * The schema will be stored in the `table_schema` attribute of the output.
    * `table_schema` is used as the attribute name to avoid conflicts with the `schema` attribute of Pydantic's
        BaseModel.

    Example
    -------
    ```python
    schema = (
        GetTableSchema(
            database="MY_DB",
            schema_="MY_SCHEMA",
            warehouse="MY_WH",
            user="gid.account@nike.com",
            password="super-secret-password",
            role="APPLICATION.SNOWFLAKE.ADMIN",
            table="MY_TABLE",
        )
        .execute()
        .table_schema
    )
    ```
    """

    table: str = Field(default=..., description="The Snowflake table name")

    class Output(StepOutput):
        """Output class for GetTableSchema"""

        table_schema: t.StructType = Field(default=..., serialization_alias="schema", description="The Spark Schema")

    def execute(self) -> Output:
        query = f"SELECT * FROM {self.table} LIMIT 1"  # nosec B608: hardcoded_sql_expressions
        df = Query(**self.get_options(), query=query).execute().df
        self.output.table_schema = df.schema


class AddColumn(SnowflakeStep):
    """
    Add an empty column to a Snowflake table with given name and DataType

    Example
    -------
    ```python
    AddColumn(
        database="MY_DB",
        schema_="MY_SCHEMA",
        warehouse="MY_WH",
        user="gid.account@nike.com",
        password=Secret("super-secret-password"),
        role="APPLICATION.SNOWFLAKE.ADMIN",
        table="MY_TABLE",
        col="MY_COL",
        dataType=StringType(),
    ).execute()
    ```
    """

    table: str = Field(default=..., description="The name of the Snowflake table")
    column: str = Field(default=..., description="The name of the new column")
    type: f.DataType = Field(default=..., description="The DataType represented as a Spark DataType")

    class Output(SnowflakeStep.Output):
        """Output class for AddColumn"""

        query: str = Field(default=..., description="Query that was executed to add the column")

    def execute(self):
        query = f"ALTER TABLE {self.table} ADD COLUMN {self.column} {map_spark_type(self.type)}".upper()
        self.output.query = query
        RunQuery(**self.get_options(), query=query).execute()


class SyncTableAndDataFrameSchema(SnowflakeStep, SnowflakeTransformation):
    """
    Sync the schema's of a Snowflake table and a DataFrame. This will add NULL columns for the columns that are not in
    both and perform type casts where needed.

    The Snowflake table will take priority in case of type conflicts.
    """

    df: DataFrame = Field(default=..., description="The Spark DataFrame")
    table: str = Field(default=..., description="The table name")
    dry_run: Optional[bool] = Field(default=False, description="Only show schema differences, do not apply changes")

    class Output(SparkStep.Output):
        """Output class for SyncTableAndDataFrameSchema"""

        original_df_schema: t.StructType = Field(default=..., description="Original DataFrame schema")
        original_sf_schema: t.StructType = Field(default=..., description="Original Snowflake schema")
        new_df_schema: t.StructType = Field(default=..., description="New DataFrame schema")
        new_sf_schema: t.StructType = Field(default=..., description="New Snowflake schema")
        sf_table_altered: bool = Field(
            default=False, description="Flag to indicate whether Snowflake schema has been altered"
        )

    def execute(self):
        self.log.warning("Snowflake table will always take a priority in case of data type conflicts!")

        # spark side
        df_schema = self.df.schema
        self.output.original_df_schema = deepcopy(df_schema)  # using deepcopy to avoid storing in place changes
        df_cols = [c.name.lower() for c in df_schema]

        # snowflake side
        sf_schema = GetTableSchema(**self.get_options(), table=self.table).execute().table_schema
        self.output.original_sf_schema = sf_schema
        sf_cols = [c.name.lower() for c in sf_schema]

        if self.dry_run:
            # Display differences between Spark DataFrame and Snowflake schemas
            # and provide dummy values that are expected as class outputs.
            self.log.warning(f"Columns to be added to Snowflake table: {set(df_cols) - set(sf_cols)}")
            self.log.warning(f"Columns to be added to Spark DataFrame: {set(sf_cols) - set(df_cols)}")

            self.output.new_df_schema = t.StructType()
            self.output.new_sf_schema = t.StructType()
            self.output.df = self.df
            self.output.sf_table_altered = False

        else:
            # Add columns to SnowFlake table that exist in DataFrame
            for df_column in df_schema:
                if df_column.name.lower() not in sf_cols:
                    AddColumn(
                        **self.get_options(),
                        table=self.table,
                        column=df_column.name,
                        type=df_column.dataType,
                    ).execute()
                    self.output.sf_table_altered = True

            if self.output.sf_table_altered:
                sf_schema = GetTableSchema(**self.get_options(), table=self.table).execute().table_schema
                sf_cols = [c.name.lower() for c in sf_schema]

            self.output.new_sf_schema = sf_schema

            # Add NULL columns to the DataFrame if they exist in SnowFlake but not in the df
            df = self.df
            for sf_col in self.output.original_sf_schema:
                sf_col_name = sf_col.name.lower()
                if sf_col_name not in df_cols:
                    sf_col_type = sf_col.dataType
                    df = df.withColumn(sf_col_name, f.lit(None).cast(sf_col_type))

            # Put DataFrame columns in the same order as the Snowflake table
            df = df.select(*sf_cols)

            self.output.df = df
            self.output.new_df_schema = df.schema


class SnowflakeWriter(SnowflakeBaseModel, Writer):
    """Class for writing to Snowflake

    See Also
    --------
    - [koheesio.steps.writers.Writer](writers/index.md#koheesio.spark.writers.Writer)
    - [koheesio.steps.writers.BatchOutputMode](writers/index.md#koheesio.spark.writers.BatchOutputMode)
    - [koheesio.steps.writers.StreamingOutputMode](writers/index.md#koheesio.spark.writers.StreamingOutputMode)
    """

    table: str = Field(default=..., description="Target table name")
    insert_type: Optional[BatchOutputMode] = Field(
        BatchOutputMode.APPEND, alias="mode", description="The insertion type, append or overwrite"
    )

    def execute(self):
        """Write to Snowflake"""
        self.log.debug(f"writing to {self.table} with mode {self.insert_type}")
        self.df.write.format(self.format).options(**self.get_options()).option("dbtable", self.table).mode(
            self.insert_type
        ).save()


class TagSnowflakeQuery(Step, ExtraParamsMixin):
    """
    Provides Snowflake query tag pre-action that can be used to easily find queries through SF history search
    and further group them for debugging and cost tracking purposes.

    Takes in query tag attributes as kwargs and additional Snowflake options dict that can optionally contain
    other set of pre-actions to be applied to a query, in that case existing pre-action aren't dropped, query tag
    pre-action will be added to them.

    Passed Snowflake options dictionary is not modified in-place, instead anew dictionary containing updated pre-actions
    is returned.

    Notes
    -----
    See this article for explanation: https://select.dev/posts/snowflake-query-tags

    Arbitrary tags can be applied, such as team, dataset names, business capability, etc.

    Example
    -------
    ```python
    query_tag = AddQueryTag(
        options={"preactions": ...},
        task_name="cleanse_task",
        pipeline_name="ingestion-pipeline",
        etl_date="2022-01-01",
        pipeline_execution_time="2022-01-01T00:00:00",
        task_execution_time="2022-01-01T01:00:00",
        environment="dev",
        trace_id="e0fdec43-a045-46e5-9705-acd4f3f96045",
        span_id="cb89abea-1c12-471f-8b12-546d2d66f6cb",
        ),
    ).execute().options
    ```
    """

    options: Dict = Field(
        default_factory=dict, description="Additional Snowflake options, optionally containing additional preactions"
    )

    class Output(StepOutput):
        """Output class for AddQueryTag"""

        options: Dict = Field(default=..., description="Copy of provided SF options, with added query tag preaction")

    def execute(self):
        """Add query tag preaction to Snowflake options"""
        tag_json = json.dumps(self.extra_params, indent=4, sort_keys=True)
        tag_preaction = f"ALTER SESSION SET QUERY_TAG = '{tag_json}';"
        preactions = self.options.get("preactions", "")
        preactions = f"{preactions}\n{tag_preaction}".strip()
        updated_options = dict(self.options)
        updated_options["preactions"] = preactions
        self.output.options = updated_options


class SynchronizeDeltaToSnowflakeTask(SnowflakeStep):
    """
    Synchronize a Delta table to a Snowflake table

    * Overwrite - only in batch mode
    * Append - supports batch and streaming mode
    * Merge - only in streaming mode

    Example
    -------
    ```python
    SynchronizeDeltaToSnowflakeTask(
        url="acme.snowflakecomputing.com",
        user="admin",
        role="ADMIN",
        warehouse="SF_WAREHOUSE",
        database="SF_DATABASE",
        schema="SF_SCHEMA",
        source_table=DeltaTableStep(...),
        target_table="my_sf_table",
        key_columns=[
            "id",
        ],
        streaming=False,
    ).run()
    ```
    """

    source_table: DeltaTableStep = Field(default=..., description="Source delta table to synchronize")
    target_table: str = Field(default=..., description="Target table in snowflake to synchronize to")
    synchronisation_mode: BatchOutputMode = Field(
        default=BatchOutputMode.MERGE,
        description="Determines if synchronisation will 'overwrite' any existing table, 'append' new rows or "
        "'merge' with existing rows.",
    )
    checkpoint_location: Optional[str] = Field(default=None, description="Checkpoint location to use")
    schema_tracking_location: Optional[str] = Field(
        default=None,
        description="Schema tracking location to use. "
        "Info: https://docs.delta.io/latest/delta-streaming.html#-schema-tracking",
    )
    staging_table_name: Optional[str] = Field(
        default=None, alias="staging_table", description="Optional snowflake staging name", validate_default=False
    )
    key_columns: Optional[List[str]] = Field(
        default_factory=list,
        description="Key columns on which merge statements will be MERGE statement will be applied.",
    )
    streaming: Optional[bool] = Field(
        default=False,
        description="Should synchronisation happen in streaming or in batch mode. Streaming is supported in 'APPEND' "
        "and 'MERGE' mode. Batch is supported in 'OVERWRITE' and 'APPEND' mode.",
    )
    persist_staging: Optional[bool] = Field(
        default=False,
        description="In case of debugging, set `persist_staging` to True to retain the staging table for inspection "
        "after synchronization.",
    )

    enable_deletion: Optional[bool] = Field(
        default=False,
        description="In case of merge synchronisation_mode add deletion statement in merge query.",
    )

    writer_: Optional[Union[ForEachBatchStreamWriter, SnowflakeWriter]] = None

    @field_validator("staging_table_name")
    def _validate_staging_table(cls, staging_table_name):
        """Validate the staging table name and return it if it's valid."""
        if "." in staging_table_name:
            raise ValueError(
                "Custom staging table must not contain '.', it is located in the same Schema as the target table."
            )
        return staging_table_name

    @model_validator(mode="before")
    def _checkpoint_location_check(cls, values: Dict):
        """Give a warning if checkpoint location is given but not expected and vice versa"""
        streaming = values.get("streaming")
        checkpoint_location = values.get("checkpoint_location")
        log = LoggingFactory.get_logger(cls.__name__)

        if streaming is False and checkpoint_location is not None:
            log.warning("checkpoint_location is provided but will be ignored in batch mode")
        if streaming is True and checkpoint_location is None:
            log.warning("checkpoint_location is not provided in streaming mode")
        return values

    @model_validator(mode="before")
    def _synch_mode_check(cls, values: Dict):
        """Validate requirements for various synchronisation modes"""
        streaming = values.get("streaming")
        synchronisation_mode = values.get("synchronisation_mode")
        key_columns = values.get("key_columns")

        allowed_output_modes = [BatchOutputMode.OVERWRITE, BatchOutputMode.MERGE, BatchOutputMode.APPEND]

        if synchronisation_mode not in allowed_output_modes:
            raise ValueError(
                f"Synchronisation mode should be one of {', '.join([m.value for m in allowed_output_modes])}"
            )
        if synchronisation_mode == BatchOutputMode.OVERWRITE and streaming is True:
            raise ValueError("Synchronisation mode can't be 'OVERWRITE' with streaming enabled")
        if synchronisation_mode == BatchOutputMode.MERGE and streaming is False:
            raise ValueError("Synchronisation mode can't be 'MERGE' with streaming disabled")
        if synchronisation_mode == BatchOutputMode.MERGE and len(key_columns) < 1:
            raise ValueError("MERGE synchronisation mode requires a list of PK columns in `key_columns`.")

        return values

    @property
    def non_key_columns(self) -> List[str]:
        """Columns of source table that aren't part of the (composite) primary key"""
        lowercase_key_columns: Set[str] = {c.lower() for c in self.key_columns}
        source_table_columns = self.source_table.columns
        non_key_columns: List[str] = [c for c in source_table_columns if c.lower() not in lowercase_key_columns]
        return non_key_columns

    @property
    def staging_table(self):
        """Intermediate table on snowflake where staging results are stored"""
        if stg_tbl_name := self.staging_table_name:
            return stg_tbl_name

        return f"{self.source_table.table}_stg"

    @property
    def reader(self):
        """
        DeltaTable reader

        Returns:
        --------
            DeltaTableReader the will yield source delta table
        """
        # Wrap in lambda functions to mimic lazy evaluation.
        # This ensures the Task doesn't fail if a config isn't provided for a reader/writer that isn't used anyway
        map_mode_reader = {
            BatchOutputMode.OVERWRITE: lambda: DeltaTableReader(
                table=self.source_table, streaming=False, schema_tracking_location=self.schema_tracking_location
            ),
            BatchOutputMode.APPEND: lambda: DeltaTableReader(
                table=self.source_table,
                streaming=self.streaming,
                schema_tracking_location=self.schema_tracking_location,
            ),
            BatchOutputMode.MERGE: lambda: DeltaTableStreamReader(
                table=self.source_table, read_change_feed=True, schema_tracking_location=self.schema_tracking_location
            ),
        }
        return map_mode_reader[self.synchronisation_mode]()

    def _get_writer(self) -> Union[SnowflakeWriter, ForEachBatchStreamWriter]:
        """
        Writer to persist to snowflake

        Depending on configured options, this returns an SnowflakeWriter or ForEachBatchStreamWriter:
        - OVERWRITE/APPEND mode yields SnowflakeWriter
        - MERGE mode yields ForEachBatchStreamWriter

        Returns
        -------
        ForEachBatchStreamWriter | SnowflakeWriter
            The right writer for the configured options and mode
        """
        # Wrap in lambda functions to mimic lazy evaluation.
        # This ensures the Task doesn't fail if a config isn't provided for a reader/writer that isn't used anyway
        map_mode_writer = {
            (BatchOutputMode.OVERWRITE, False): lambda: SnowflakeWriter(
                table=self.target_table, insert_type=BatchOutputMode.OVERWRITE, **self.get_options()
            ),
            (BatchOutputMode.APPEND, False): lambda: SnowflakeWriter(
                table=self.target_table, insert_type=BatchOutputMode.APPEND, **self.get_options()
            ),
            (BatchOutputMode.APPEND, True): lambda: ForEachBatchStreamWriter(
                checkpointLocation=self.checkpoint_location,
                batch_function=writer_to_foreachbatch(
                    SnowflakeWriter(table=self.target_table, insert_type=BatchOutputMode.APPEND, **self.get_options())
                ),
            ),
            (BatchOutputMode.MERGE, True): lambda: ForEachBatchStreamWriter(
                checkpointLocation=self.checkpoint_location,
                batch_function=self._merge_batch_write_fn(
                    key_columns=self.key_columns,
                    non_key_columns=self.non_key_columns,
                    staging_table=self.staging_table,
                ),
            ),
        }
        return map_mode_writer[(self.synchronisation_mode, self.streaming)]()

    @property
    def writer(self) -> Union[ForEachBatchStreamWriter, SnowflakeWriter]:
        """
        Writer to persist to snowflake

        Depending on configured options, this returns an SnowflakeWriter or ForEachBatchStreamWriter:
        - OVERWRITE/APPEND mode yields SnowflakeWriter
        - MERGE mode yields ForEachBatchStreamWriter

        Returns
        -------
            Union[ForEachBatchStreamWriter, SnowflakeWriter]
        """
        # Cache 'writer' object in memory to ensure same object is used everywhere, this ensures access to underlying
        # member objects such as active streaming queries (if any).
        if not self.writer_:
            self.writer_ = self._get_writer()
        return self.writer_

    def truncate_table(self, snowflake_table):
        """Truncate a given snowflake table"""
        truncate_query = f"""TRUNCATE TABLE IF EXISTS {snowflake_table}"""
        query_executor = RunQuery(
            **self.get_options(),
            query=truncate_query,
        )
        query_executor.execute()

    def drop_table(self, snowflake_table):
        """Drop a given snowflake table"""
        self.log.warning(f"Dropping table {snowflake_table} from snowflake")
        drop_table_query = f"""DROP TABLE IF EXISTS {snowflake_table}"""
        query_executor = RunQuery(**self.get_options(), query=drop_table_query)
        query_executor.execute()

    def _merge_batch_write_fn(self, key_columns, non_key_columns, staging_table):
        """Build a batch write function for merge mode"""

        # pylint: disable=unused-argument
        def inner(dataframe: DataFrame, batchId: int):
            self._build_staging_table(dataframe, key_columns, non_key_columns, staging_table)
            self._merge_staging_table_into_target()

        # pylint: enable=unused-argument
        return inner

    @staticmethod
    def _compute_latest_changes_per_pk(
        dataframe: DataFrame, key_columns: List[str], non_key_columns: List[str]
    ) -> DataFrame:
        """Compute the latest changes per primary key"""
        windowSpec = Window.partitionBy(*key_columns).orderBy(f.col("_commit_version").desc())
        ranked_df = (
            dataframe.filter("_change_type != 'update_preimage'")
            .withColumn("rank", f.rank().over(windowSpec))
            .filter("rank = 1")
            .select(*key_columns, *non_key_columns, "_change_type")  # discard unused columns
            .distinct()
        )
        return ranked_df

    def _build_staging_table(self, dataframe, key_columns, non_key_columns, staging_table):
        """Build snowflake staging table"""
        ranked_df = self._compute_latest_changes_per_pk(dataframe, key_columns, non_key_columns)
        batch_writer = SnowflakeWriter(
            table=staging_table, df=ranked_df, insert_type=BatchOutputMode.APPEND, **self.get_options()
        )
        batch_writer.execute()

    def _merge_staging_table_into_target(self) -> None:
        """
        Merge snowflake staging table into final snowflake table
        """
        merge_query = self._build_sf_merge_query(
            target_table=self.target_table,
            stage_table=self.staging_table,
            pk_columns=self.key_columns,
            non_pk_columns=self.non_key_columns,
            enable_deletion=self.enable_deletion,
        )

        query_executor = RunQuery(
            **self.get_options(),
            query=merge_query,
        )
        query_executor.execute()

    @staticmethod
    def _build_sf_merge_query(
        target_table: str, stage_table: str, pk_columns: List[str], non_pk_columns, enable_deletion: bool = False
    ):
        """Build a CDF merge query string

        Parameters
        ----------
        target_table: Table
            Destination table to merge into
        stage_table: Table
            Temporary table containing updates to be executed
        pk_columns: List[str]
            Column names used to uniquely identify each row
        non_pk_columns: List[str]
            Non-key columns that may need to be inserted/updated
        enable_deletion: bool
            DELETE actions are synced. If set to False (default) then sync is non-destructive

        Returns
        -------
        str
            Query to be executed on the target database
        """
        all_fields = [*pk_columns, *non_pk_columns]
        key_join_string = " AND ".join(f"target.{k} = temp.{k}" for k in pk_columns)
        columns_string = ", ".join(all_fields)
        assignment_string = ", ".join(f"{k} = temp.{k}" for k in non_pk_columns)
        values_string = ", ".join(f"temp.{k}" for k in all_fields)

        query = f"""
        MERGE INTO {target_table} target
        USING {stage_table} temp ON {key_join_string}
        WHEN MATCHED AND temp._change_type = 'update_postimage' THEN UPDATE SET {assignment_string}
        WHEN NOT MATCHED AND temp._change_type != 'delete' THEN INSERT ({columns_string}) VALUES ({values_string})
        """  # nosec B608: hardcoded_sql_expressions
        if enable_deletion:
            query += "WHEN MATCHED AND temp._change_type = 'delete' THEN DELETE"

        return query

    def extract(self) -> DataFrame:
        """
        Extract source table
        """
        if self.synchronisation_mode == BatchOutputMode.MERGE:
            if not self.source_table.is_cdf_active:
                raise RuntimeError(
                    f"Source table {self.source_table.table_name} does not have CDF enabled. "
                    f"Set TBLPROPERTIES ('delta.enableChangeDataFeed' = true) to enable. "
                    f"Current properties = {self.source_table_properties}"
                )

        df = self.reader.read()
        self.output.source_df = df
        return df

    def load(self, df) -> DataFrame:
        """Load source table into snowflake"""
        if self.synchronisation_mode == BatchOutputMode.MERGE:
            self.log.info(f"Truncating staging table {self.staging_table}")
            self.truncate_table(self.staging_table)
        self.writer.write(df)
        self.output.target_df = df
        return df

    def execute(self) -> None:
        # extract
        df = self.extract()
        self.output.source_df = df

        # synchronize
        self.output.target_df = df
        self.load(df)
        if not self.persist_staging:
            # If it's a streaming job, await for termination before dropping staging table
            if self.streaming:
                self.writer.await_termination()
            self.drop_table(self.staging_table)

    def run(self):
        """alias of execute"""
        return self.execute()



"""
A collection of classes for performing various transformations on arrays in PySpark.

These transformations include operations such as removing duplicates, exploding arrays into separate rows, reversing
the order of elements, sorting elements, removing certain values, and calculating aggregate statistics like minimum,
maximum, sum, mean, and median.

Concept
-------
* Every transformation in this module is implemented as a class that inherits from the `ArrayTransformation` class.
* The `ArrayTransformation` class is a subclass of `ColumnsTransformationWithTarget`
* The `ArrayTransformation` class implements the `func` method, which is used to define the transformation logic.
* The `func` method takes a `column` as input and returns a `Column` object.
* The `Column` object is a PySpark column that can be used to perform transformations on a DataFrame column.
* The `ArrayTransformation` limits the data type of the transformation to array by setting the `ColumnConfig` class to
    `run_for_all_data_type = [SparkDatatype.ARRAY]` and `limit_data_type = [SparkDatatype.ARRAY]`.

See Also
--------
* [koheesio.spark.transformations](index.md)
    Module containing all transformation classes.
* [koheesio.spark.transformations.ColumnsTransformationWithTarget](index.md#koheesio.spark.transformations.ColumnsTransformationWithTarget)
    Base class for all transformations that operate on columns and have a target column.
"""

from typing import Any
from abc import ABC
from functools import reduce

from pyspark.sql import Column
from pyspark.sql import functions as F

from koheesio.models import Field
from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils import (
    SparkDatatype,
    spark_data_type_is_numeric,
    spark_minor_version,
)

__all__ = [
    "ArrayDistinct",
    "Explode",
    "ExplodeDistinct",
    "ArrayReverse",
    "ArraySort",
    "ArraySortAsc",
    "ArraySortDesc",
    "ArrayRemove",
    "ArrayMin",
    "ArrayMax",
    "ArraySum",
    "ArrayMean",
    "ArrayMedian",
]


class ArrayTransformation(ColumnsTransformationWithTarget, ABC):
    """Base class for array transformations"""

    # pylint: disable=R0903
    class ColumnConfig(ColumnsTransformationWithTarget.ColumnConfig):
        """Set the data type of the Transformation to array"""

        run_for_all_data_type = [SparkDatatype.ARRAY]
        limit_data_type = [SparkDatatype.ARRAY]

    # pylint: enable=R0903

    def func(self, column: Column) -> Column:
        raise NotImplementedError("This is an abstract class")


class ArrayDistinct(ArrayTransformation):
    """
    Remove duplicates from array

    Example
    -------
    ```python
    ArrayDistinct(column="array_column")
    ```
    """

    filter_empty: bool = Field(
        default=True, description="Remove null, nan, and empty values from array. Default is True."
    )

    def func(self, column: Column) -> Column:
        _fn = F.array_distinct(column)

        # noinspection PyUnresolvedReferences
        element_type = self.column_type_of_col(column, None, False).elementType
        is_numeric = spark_data_type_is_numeric(element_type)

        if self.filter_empty:
            # Remove null values from array
            if spark_minor_version >= 3.4:
                # Run array_compact if spark version is 3.4 or higher
                # https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.array_compact.html
                # pylint: disable=E0611
                from pyspark.sql.functions import array_compact as _array_compact

                _fn = _array_compact(_fn)
                # pylint: enable=E0611
            else:
                # Otherwise, remove null from array using array_except
                _fn = F.array_except(_fn, F.array(F.lit(None)))

            # Remove nan or empty values from array (depends on the type of the elements in array)
            if is_numeric:
                # Remove nan from array (float/int/numbers)
                _fn = F.array_except(_fn, F.array(F.lit(float("nan")).cast(element_type)))
            else:
                # Remove empty values from array (string/text)
                _fn = F.array_except(_fn, F.array(F.lit(""), F.lit(" ")))

        return _fn


class Explode(ArrayTransformation):
    """
    Explode the array into separate rows

    Example
    -------
    ```python
    Explode(column="array_column")
    ```
    """

    distinct: bool = Field(False, description="Remove duplicates from the exploded array. Default is False.")
    preserve_nulls: bool = Field(
        True,
        description="Preserve rows with null values in the exploded array"
        " by using explode_outer instead of explode.Default is True.",
    )

    def func(self, column: Column) -> Column:
        if self.distinct:
            column = ArrayDistinct.from_step(self).func(column)
        return F.explode_outer(column) if self.preserve_nulls else F.explode(column)


class ExplodeDistinct(Explode):
    """
    Explode the array into separate rows while removing duplicates and empty values

    Example
    -------
    ```python
    ExplodeDistinct(column="array_column")
    ```
    """

    distinct: bool = True


class ArrayReverse(ArrayTransformation):
    """
    Reverse the order of elements in the array

    Example
    -------
    ```python
    ArrayReverse(column="array_column")
    ```
    """

    def func(self, column: Column) -> Column:
        return F.reverse(column)


class ArraySort(ArrayTransformation):
    """
    Sort the elements in the array

    By default, the elements are sorted in ascending order. To sort the elements in descending order, set the `reverse`
    parameter to True.

    Example
    -------
    ```python
    ArraySort(column="array_column")
    ```
    """

    reverse: bool = Field(
        default=False, description="Sort the elements in the array in a descending order. Default is False."
    )

    def func(self, column: Column) -> Column:
        column = F.array_sort(column)
        if self.reverse:
            # Reverse the order of elements in the array
            column = ArrayReverse.from_step(self).func(column)
        return column


ArraySortAsc = ArraySort


class ArraySortDesc(ArraySort):
    """Sort the elements in the array in descending order"""

    reverse: bool = True


class ArrayNullNanProcess(ArrayTransformation):
    """
    Process an array by removing NaN and/or NULL values from elements.

    Parameters
    ----------
    keep_nan : bool, default False
        Whether to keep NaN values in the array. If set to True, the NaN values will be kept in the array.

    keep_null : bool, default False
        Whether to keep NULL values in the array. If set to True, the NULL values will be kept in the array.

    Returns
    -------
    column : Column
        The processed column with NaN and/or NULL values removed from elements.

    Examples
    --------
    ```python
    >>> input_data = [(1, [1.1, 2.1, 4.1, float("nan")])]
    >>> input_schema = StructType([StructField("id", IntegerType(), True),
        StructField("array_float", ArrayType(FloatType()), True),
    ])
    >>> spark = SparkSession.builder.getOrCreate()
    >>> df = spark.createDataFrame(input_data, schema=input_schema)
    >>> transformer = ArrayNumericNanProcess(column="array_float", keep_nan=False)
    >>> transformer.transform(df)
    >>> print(transformer.output.df.collect()[0].asDict()["array_float"])
    [1.1, 2.1, 4.1]

    >>> input_data = [(1, [1.1, 2.2, 4.1, float("nan")])]
    >>> input_schema = StructType([StructField("id", IntegerType(), True),
        StructField("array_float", ArrayType(FloatType()), True),
    ])
    >>> spark = SparkSession.builder.getOrCreate()
    >>> df = spark.createDataFrame(input_data, schema=input_schema)
    >>> transformer = ArrayNumericNanProcess(column="array_float", keep_nan=True)
    >>> transformer.transform(df)
    >>> print(transformer.output.df.collect()[0].asDict()["array_float"])
    [1.1, 2.1, 4.1, nan]
    ```
    """

    keep_nan: bool = Field(
        False,
        description="Whether to keep nan values in the array. Default is False. "
        "If set to True, the nan values will be kept in the array.",
    )

    keep_null: bool = Field(
        False,
        description="Whether to keep null values in the array. Default is False. "
        "If set to True, the null values will be kept in the array.",
    )

    def func(self, column: Column) -> Column:
        """
        Process the given column by removing NaN and/or NULL values from elements.

        Parameters:
        -----------
        column : Column
            The column to be processed.

        Returns:
        --------
        column : Column
            The processed column with NaN and/or NULL values removed from elements.
        """

        def apply_logic(x: Column):
            if self.keep_nan is False and self.keep_null is False:
                logic = x.isNotNull() & ~F.isnan(x)
            elif self.keep_nan is False:
                logic = ~F.isnan(x)
            elif self.keep_null is False:
                logic = x.isNotNull()

            return logic

        if self.keep_nan is False or self.keep_null is False:
            column = F.filter(column, apply_logic)

        return column


class ArrayRemove(ArrayNullNanProcess):
    """
    Remove a certain value from the array

    Parameters
    ----------
    keep_nan : bool, default False
        Whether to keep NaN values in the array. If set to True, the NaN values will be kept in the array.

    keep_null : bool, default False
        Whether to keep NULL values in the array. If set to True, the NULL values will be kept in the array.

    Example
    -------
    ```python
    ArrayRemove(column="array_column", value="value_to_remove")
    ```
    """

    value: Any = Field(default=None, description="The value to remove from the array.")
    make_distinct: bool = Field(default=False, description="Whether to remove duplicates from the array.")

    def func(self, column: Column) -> Column:
        value = self.value

        column = super().func(column)

        def filter_logic(x: Column, _val: Any):
            if self.keep_null and self.keep_nan:
                logic = (x != F.lit(_val)) | x.isNull() | F.isnan(x)
            elif self.keep_null:
                logic = (x != F.lit(_val)) | x.isNull()
            elif self.keep_nan:
                logic = (x != F.lit(_val)) | F.isnan(x)
            else:
                logic = x != F.lit(_val)

            return logic

        # Check if the value is iterable (i.e., a list, tuple, or set)
        if isinstance(value, (list, tuple, set)):
            result = reduce(lambda res, val: F.filter(res, lambda x: filter_logic(x, val)), value, column)
        else:
            # If the value is not iterable, simply remove the value from the array
            result = F.filter(column, lambda x: filter_logic(x, value))

        if self.make_distinct:
            result = F.array_distinct(result)

        return result


class ArrayMin(ArrayTransformation):
    """
    Return the minimum value in the array

    Example
    -------
    ```python
    ArrayMin(column="array_column")
    ```
    """

    def func(self, column: Column) -> Column:
        return F.array_min(column)


class ArrayMax(ArrayNullNanProcess):
    """
    Return the maximum value in the array

    Example
    -------
    ```python
    ArrayMax(column="array_column")
    ```
    """

    def func(self, column: Column) -> Column:
        # Call for processing of nan values
        column = super().func(column)

        return F.array_max(column)


class ArraySum(ArrayNullNanProcess):
    """
    Return the sum of the values in the array

    Parameters
    ----------
    keep_nan : bool, default False
        Whether to keep NaN values in the array. If set to True, the NaN values will be kept in the array.

    keep_null : bool, default False
        Whether to keep NULL values in the array. If set to True, the NULL values will be kept in the array.

    Example
    -------
    ```python
    ArraySum(column="array_column")
    ```
    """

    def func(self, column: Column) -> Column:
        """Using the `aggregate` function to sum the values in the array"""
        # raise an error if the array contains non-numeric elements
        element_type = self.column_type_of_col(column, None, False).elementType
        if not spark_data_type_is_numeric(element_type):
            raise ValueError(
                f"{column = } contains non-numeric values. The array type is {element_type}. "
                f"Only numeric values are supported for summing."
            )

        # remove na values from array.
        column = super().func(column)

        # Using the `aggregate` function to sum the values in the array by providing the initial value as 0.0 and the
        # lambda function to add the elements together. Pyspark will automatically infer the type of the initial value
        # making 0.0 valid for both integer and float types.
        initial_value = F.lit(0.0)
        return F.aggregate(column, initial_value, lambda accumulator, x: accumulator + x)


class ArrayMean(ArrayNullNanProcess):
    """
    Return the mean of the values in the array.

    Note: Only numeric values are supported for calculating the mean.

    Example
    -------
    ```python
    ArrayMean(column="array_column", target_column="average")
    ```
    """

    def func(self, column: Column) -> Column:
        """Calculate the mean of the values in the array"""
        # raise an error if the array contains non-numeric elements
        element_type = self.column_type_of_col(col=column, df=None, simple_return_mode=False).elementType

        if not spark_data_type_is_numeric(element_type):
            raise ValueError(
                f"{column = } contains non-numeric values. The array type is {element_type}. "
                f"Only numeric values are supported for calculating a mean."
            )

        _sum = ArraySum.from_step(self).func(column)
        # Call for processing of nan values
        column = super().func(column)
        _size = F.size(column)
        # return 0 if the size of the array is 0 to avoid division by zero
        return F.when(_size == 0, F.lit(0)).otherwise(_sum / _size)


class ArrayMedian(ArrayNullNanProcess):
    """
    Return the median of the values in the array.

    The median is the middle value in a sorted, ascending or descending, list of numbers.

    - If the size of the array is even, the median is the average of the two middle numbers.
    - If the size of the array is odd, the median is the middle number.

    Note: Only numeric values are supported for calculating the median.

    Example
    -------
    ```python
    ArrayMedian(column="array_column", target_column="median")
    ```
    """

    def func(self, column: Column) -> Column:
        """Calculate the median of the values in the array"""
        # Call for processing of nan values
        column = super().func(column)

        sorted_array = ArraySort.from_step(self).func(column)
        _size: Column = F.size(sorted_array)

        # Calculate the middle index. If the size is odd, PySpark discards the fractional part.
        # Use floor function to ensure the result is an integer
        middle: Column = F.floor((_size + 1) / 2).cast("int")

        # Define conditions
        is_size_zero: Column = _size == 0
        is_column_null: Column = column.isNull()
        is_size_even: Column = _size % 2 == 0

        # Define actions / responses
        # For even-sized arrays, calculate the average of the two middle elements
        average_of_middle_elements = (F.element_at(sorted_array, middle) + F.element_at(sorted_array, middle + 1)) / 2
        # For odd-sized arrays, select the middle element
        middle_element = F.element_at(sorted_array, middle)
        # In case the array is empty, return either None or 0
        none_value = F.lit(None)
        zero_value = F.lit(0)

        median = (
            # Check if the size of the array is 0
            F.when(
                is_size_zero,
                # If the size of the array is 0 and the column is null, return None
                # If the size of the array is 0 and the column is not null, return 0
                F.when(is_column_null, none_value).otherwise(zero_value),
            ).otherwise(
                # If the size of the array is not 0, calculate the median
                F.when(is_size_even, average_of_middle_elements).otherwise(middle_element)
            )
        )

        return median



"""Module that holds some classes and functions to be able to write to a stream

Classes
-------
Trigger
    class to set the trigger for a stream query
StreamWriter
    abstract class for stream writers
ForEachBatchStreamWriter
    class to run a writer for each batch

Functions
--------
writer_to_foreachbatch
    function to be used as batch_function for StreamWriter (sub)classes
"""

from typing import Callable, Dict, Optional, Union
from abc import ABC, abstractmethod

from pyspark.sql.streaming import DataStreamWriter, StreamingQuery

from koheesio import Step
from koheesio.models import ConfigDict, Field, field_validator, model_validator
from koheesio.spark.writers import StreamingOutputMode, Writer
from koheesio.utils import convert_str_to_bool


class Trigger(Step):
    """Trigger types for a stream query.

    Only one trigger can be set!

    Example
    -------

    - processingTime='5 seconds'
    - continuous='5 seconds'
    - availableNow=True
    - once=True

    See Also
    --------
    - https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#triggers
    """

    processing_time: Optional[str] = Field(
        default=None,
        alias="processingTime",
        description="a processing time interval as a string, e.g. '5 seconds', '1 minute'."
        "Set a trigger that runs a microbatch query periodically based on the processing time.",
    )
    continuous: Optional[str] = Field(
        default=None,
        description="a time interval as a string, e.g. '5 seconds', '1 minute'."
        "Set a trigger that runs a continuous query with a given checkpoint interval.",
    )
    once: Optional[bool] = Field(
        default=None,
        deprecated=True,
        description="if set to True, set a trigger that processes only one batch of data in a streaming query then "
        "terminates the query. use `available_now` instead of `once`.",
    )
    available_now: Optional[bool] = Field(
        default=None,
        alias="availableNow",
        description="if set to True, set a trigger that processes all available data in multiple batches then "
        "terminates the query.",
    )

    model_config = ConfigDict(validate_default=False, extra="forbid")

    @classmethod
    def _all_triggers_with_alias(cls):
        """Internal method to return all trigger types with their alias. Used for logging purposes"""
        fields = cls.model_fields
        triggers = [
            f"{field_info.alias}/{field}" if field_info.alias else field
            for field, field_info in fields.items()
            if field not in ["name", "description"]
        ]
        return triggers

    @property
    def triggers(self):
        """Returns a list of tuples with the value for each trigger"""
        return self.model_dump(exclude={"name", "description"}, by_alias=True)

    @model_validator(mode="before")
    def validate_triggers(cls, triggers: Dict):
        """Validate the trigger value"""
        params = [*triggers.values()]

        # adapted from `pyspark.sql.streaming.readwriter.DataStreamWriter.trigger`; modified to work with pydantic v2
        if not triggers:
            raise ValueError("No trigger provided")
        if len(params) > 1:
            raise ValueError("Multiple triggers not allowed.")

        return triggers

    @field_validator("processing_time", mode="before")
    def validate_processing_time(cls, processing_time):
        """Validate the processing time trigger value"""
        # adapted from `pyspark.sql.streaming.readwriter.DataStreamWriter.trigger`
        if not isinstance(processing_time, str):
            raise ValueError(f"Value for processing_time must be a string. Got: {processing_time}")

        if len(processing_time.strip()) == 0:
            raise ValueError(f"Value for processingTime must be a non empty string. Got: {processing_time}")
        return processing_time

    @field_validator("continuous", mode="before")
    def validate_continuous(cls, continuous):
        """Validate the continuous trigger value"""
        # adapted from `pyspark.sql.streaming.readwriter.DataStreamWriter.trigger` except that the if statement is not
        # split in two parts
        if not isinstance(continuous, str):
            raise ValueError(f"Value for continuous must be a string. Got: {continuous}")

        if len(continuous.strip()) == 0:
            raise ValueError(f"Value for continuous must be a non empty string. Got: {continuous}")
        return continuous

    @field_validator("once", mode="before")
    def validate_once(cls, once):
        """Validate the once trigger value"""
        # making value a boolean when given
        once = convert_str_to_bool(once)

        # adapted from `pyspark.sql.streaming.readwriter.DataStreamWriter.trigger`
        if once is not True:
            raise ValueError(f"Value for once must be True. Got: {once}")
        return once

    @field_validator("available_now", mode="before")
    def validate_available_now(cls, available_now):
        """Validate the available_now trigger value"""
        # making value a boolean when given
        available_now = convert_str_to_bool(available_now)

        # adapted from `pyspark.sql.streaming.readwriter.DataStreamWriter.trigger`
        if available_now is not True:
            raise ValueError(f"Value for availableNow must be True. Got:{available_now}")
        return available_now

    @property
    def value(self) -> Dict[str, str]:
        """Returns the trigger value as a dictionary"""
        trigger = {k: v for k, v in self.triggers.items() if v is not None}
        return trigger

    @classmethod
    def from_dict(cls, _dict):
        """Creates a Trigger class based on a dictionary"""
        return cls(**_dict)

    @classmethod
    def from_string(cls, trigger: str):
        """Creates a Trigger class based on a string

        Example
        -------
        ### happy flow

        * processingTime='5 seconds'
        * processing_time="5 hours"
        * processingTime=4 minutes
        * once=True
        * once=true
        * available_now=true
        * continuous='3 hours'
        * once=TrUe
        * once=TRUE

        ### unhappy flow
        valid values, but should fail the validation check of the class

        * availableNow=False
        * continuous=True
        * once=false
        """
        import re

        trigger_from_string = re.compile(r"(?P<triggerType>\w+)=[\'\"]?(?P<value>.+)[\'\"]?")
        _match = trigger_from_string.match(trigger)

        if _match is None:
            raise ValueError(
                f"Cannot parse value for Trigger: '{trigger}'. \n"
                f"Valid types are {', '.join(cls._all_triggers_with_alias())}"
            )

        trigger_type, value = _match.groups()

        # strip the value of any quotes
        value = value.strip("'").strip('"')

        # making value a boolean when given
        value = convert_str_to_bool(value)

        return cls.from_dict({trigger_type: value})

    @classmethod
    def from_any(cls, value):
        """Dynamically creates a Trigger class based on either another Trigger class, a passed string value, or a
        dictionary

        This way Trigger.from_any can be used as part of a validator, without needing to worry about supported types
        """
        if isinstance(value, Trigger):
            return value

        if isinstance(value, str):
            return cls.from_string(value)

        if isinstance(value, dict):
            return cls.from_dict(value)

        raise RuntimeError(f"Unable to create Trigger based on the given value: {value}")

    def execute(self):
        """Returns the trigger value as a dictionary
        This method can be skipped, as the value can be accessed directly from the `value` property
        """
        self.log.warning("Trigger.execute is deprecated. Use Trigger.value directly instead")
        self.output.value = self.value


class StreamWriter(Writer, ABC):
    """ABC Stream Writer"""

    checkpoint_location: str = Field(
        default=...,
        alias="checkpointLocation",
        description="In case of a failure or intentional shutdown, you can recover the previous progress and state of "
        "a previous query, and continue where it left off. This is done using checkpointing and write-ahead logs. You "
        "can configure a query with a checkpoint location, and the query will save all the progress information and "
        "the running aggregates to the checkpoint location. This checkpoint location has to be a path in an HDFS "
        "compatible file system that is accessible by Spark (e.g. Databricks Unity Catalog External Location)",
    )

    output_mode: StreamingOutputMode = Field(
        default=StreamingOutputMode.APPEND, alias="outputMode", description=StreamingOutputMode.__doc__
    )

    batch_function: Optional[Callable] = Field(
        default=None,
        description="allows you to run custom batch functions for each micro batch",
        alias="batch_function_for_each_df",
    )

    trigger: Optional[Union[Trigger, str, Dict]] = Field(
        default=Trigger(available_now=True),
        description="Set the trigger for the stream query. If this is not set it process data as batch",
    )

    streaming_query: Optional[Union[str, StreamingQuery]] = Field(
        default=None, description="Query ID of the stream query"
    )

    @property
    def _trigger(self):
        """Returns the trigger value as a dictionary"""
        return self.trigger.value

    @field_validator("output_mode")
    def _validate_output_mode(cls, mode):
        """Ensure that the given mode is a valid StreamingOutputMode"""
        if isinstance(mode, str):
            return mode
        return str(mode.value)

    @field_validator("trigger")
    def _validate_trigger(cls, trigger):
        """Ensure that the given trigger is a valid Trigger class"""
        return Trigger.from_any(trigger)

    def await_termination(self, timeout: Optional[int] = None):
        """Await termination of the stream query"""
        self.streaming_query.awaitTermination(timeout=timeout)

    @property
    def stream_writer(self) -> DataStreamWriter:
        """Returns the stream writer for the given DataFrame and settings"""
        write_stream = self.df.writeStream.format(self.format).outputMode(self.output_mode)

        if self.checkpoint_location:
            write_stream = write_stream.option("checkpointLocation", self.checkpoint_location)

        if self.batch_function:
            write_stream = write_stream.foreachBatch(self.batch_function)

        # set trigger
        write_stream = write_stream.trigger(**self._trigger)

        return write_stream

    @property
    def writer(self):
        """Returns the stream writer since we don't have a batch mode for streams"""
        return self.stream_writer

    @abstractmethod
    def execute(self):
        raise NotImplementedError


class ForEachBatchStreamWriter(StreamWriter):
    """Runnable ForEachBatchWriter"""

    @field_validator("batch_function")
    def _validate_batch_function_exists(cls, batch_function):
        """Ensure that a batch_function is defined"""
        if not batch_function or not isinstance(batch_function, Callable):
            raise ValueError(f"{cls.__name__} requires a defined for `batch_function`")
        return batch_function

    def execute(self):
        self.streaming_query = self.writer.start()


def writer_to_foreachbatch(writer: Writer):
    """Call `writer.execute` on each batch

    To be passed as batch_function for StreamWriter (sub)classes.

    Example
    -------
    ### Writing to a Delta table and a Snowflake table
    ```python
    DeltaTableStreamWriter(
        table="my_table",
        checkpointLocation="my_checkpointlocation",
        batch_function=writer_to_foreachbatch(
            SnowflakeWriter(
                **sfOptions,
                table="snowflake_table",
                insert_type=SnowflakeWriter.InsertType.APPEND,
            )
        ),
    )
    ```
    """

    def inner(df, batch_id: int):
        """Inner method

        As per the Spark documentation:
        In every micro-batch, the provided function will be called in every micro-batch with (i) the output rows as a
        DataFrame and (ii) the batch identifier. The batchId can be used deduplicate and transactionally write the
        output (that is, the provided Dataset) to external systems. The output DataFrame is guaranteed to exactly
        same for the same batchId (assuming all operations are deterministic in the query).
        """
        writer.log.debug(f"Running batch function for batch {batch_id}")
        writer.write(df)

    return inner



import pytest
from chispa import assert_df_equality

from pyspark.sql.types import *

from koheesio.spark.readers.databricks.autoloader import AutoLoader

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "bad_format",
    [
        ("mp3"),
        ("mp4"),
    ],
)
def test_invalid_format(bad_format):
    with pytest.raises(AttributeError):
        AutoLoader(format=bad_format, location="some_path", schema_location="some_schema_location")


def mock_reader(self):
    reader = self.spark.read.format("json")
    if self.schema_ is not None:
        reader = reader.schema(self.schema_)
    return reader.options(**self.options)


def test_read_json_infer_schema(spark, mocker, data_path):
    mocker.patch("koheesio.spark.readers.databricks.autoloader.AutoLoader.reader", mock_reader)

    options = {"multiLine": "true"}

    json_file_path_str = f"{data_path}/readers/json_file/dummy.json"
    auto_loader = AutoLoader(format="json", location=json_file_path_str, schema_location="dummy_value", options=options)

    auto_loader.execute()
    result = auto_loader.output.df

    schema_expected = StructType(
        [
            StructField("string", StringType(), True),
            StructField("int", LongType(), True),
            StructField("array", ArrayType(LongType()), True),
        ]
    )

    data_expected = [
        {"string": "string1", "int": 1, "array": [1, 11, 111]},
        {"string": "string2", "int": 2, "array": [2, 22, 222]},
    ]
    expected_df = spark.createDataFrame(data_expected, schema_expected)
    assert_df_equality(result, expected_df, ignore_column_order=True)


def test_read_json_exact_explicit_schema_struct(spark, mocker, data_path):
    mocker.patch("koheesio.spark.readers.databricks.autoloader.AutoLoader.reader", mock_reader)

    schema = StructType(
        [
            StructField("string", StringType(), True),
            StructField("int", LongType(), True),
            StructField("array", ArrayType(LongType()), True),
        ]
    )
    options = {"multiLine": "true"}
    json_file_path_str = f"{data_path}/readers/json_file/dummy.json"
    auto_loader = AutoLoader(
        format="json", location=json_file_path_str, schema_location="dummy_value", options=options, schema=schema
    )

    auto_loader.execute()
    result = auto_loader.output.df

    data_expected = [
        {"string": "string1", "int": 1, "array": [1, 11, 111]},
        {"string": "string2", "int": 2, "array": [2, 22, 222]},
    ]
    expected_df = spark.createDataFrame(data_expected, schema)
    assert_df_equality(result, expected_df, ignore_column_order=True)


def test_read_json_different_explicit_schema_string(spark, mocker, data_path):
    mocker.patch("koheesio.spark.readers.databricks.autoloader.AutoLoader.reader", mock_reader)

    schema = "string STRING,array ARRAY<BIGINT>"
    options = {"multiLine": "true"}
    json_file_path_str = f"{data_path}/readers/json_file/dummy.json"
    auto_loader = AutoLoader(
        format="json", location=json_file_path_str, schema_location="dummy_value", options=options, schema=schema
    )

    auto_loader.execute()
    result = auto_loader.output.df

    data_expected = [
        {"string": "string1", "array": [1, 11, 111]},
        {"string": "string2", "array": [2, 22, 222]},
    ]
    expected_df = spark.createDataFrame(data_expected, schema)
    assert_df_equality(result, expected_df, ignore_column_order=True)
